import pytest

import datasets
import molflux.metrics
import molflux.modelzoo
from molflux.core import compute_scores, score_model

parametrized_suites = ["regression"]


@pytest.mark.parametrize("suite", parametrized_suites)
def test_can_score_model(suite):
    """That scoring a model returns scores according to standardised metrics format."""

    fold = datasets.DatasetDict(
        {
            "train": datasets.Dataset.from_dict(
                {
                    "x": [1, 2, 3],
                    "y": [3, 2, 1],
                },
            ),
            "validation": datasets.Dataset.from_dict(
                {
                    "x": [],
                    "y": [],
                },
            ),
            "test": datasets.Dataset.from_dict(
                {
                    "x": [1, 3, 2],
                    "y": [3, 1, 2],
                },
            ),
        },
    )

    model = molflux.modelzoo.load_model(
        "average_regressor",
        x_features=["x"],
        y_features=["y"],
    )
    model.train(fold["train"])

    metrics = molflux.metrics.load_suite(suite)

    scores = score_model(model, fold=fold, metrics=metrics)

    for split_name in fold.keys():
        assert split_name in scores.keys()

        if fold[split_name]:
            for y_feature in model.y_features:
                assert y_feature in scores[split_name]

                score_keys = list(scores[split_name][y_feature])

                for metric in metrics:
                    # the exact names given to score keys are not officially
                    # part of the metrics API, but for now they tend to
                    # always have metric.tag in it so do this instead of
                    # assert metric.tag in scores[split_name][y_feature]
                    assert any(metric.tag in score_key for score_key in score_keys)


def test_splits_not_in_same_order():
    """
    Computing scores should be robust to input references and predictions folds
    containing splits not in the same order.
    """

    references = datasets.DatasetDict(
        {
            "train": datasets.Dataset.from_dict(
                {
                    "x": [1, 2, 3],
                    "y": [3, 2, 1],
                },
            ),
            "validation": datasets.Dataset.from_dict(
                {
                    "x": [],
                    "y": [],
                },
            ),
            "test": datasets.Dataset.from_dict(
                {
                    "x": [1, 3, 2],
                    "y": [3, 1, 2],
                },
            ),
        },
    )

    predictions = datasets.DatasetDict(
        {
            "train": datasets.Dataset.from_dict(
                {
                    "x": [1, 2, 3],
                    "y": [3, 2, 1],
                },
            ),
            "test": datasets.Dataset.from_dict(
                {
                    "x": [1, 3, 2],
                    "y": [3, 1, 2],
                },
            ),
            "validation": datasets.Dataset.from_dict(
                {
                    "x": [],
                    "y": [],
                },
            ),
        },
    )

    metrics = molflux.metrics.load_suite("regression")

    scores = compute_scores(
        predictions=predictions,
        references=references,
        metrics=metrics,
    )

    # If scores have been computed correctly, the validation scores should be None
    # but the test scores should be actual values
    for task in scores["validation"]:
        for score in scores["validation"][task].values():
            assert score is None

    for task in scores["test"]:
        for score in scores["test"][task].values():
            assert score is not None


def test_warning_is_raised_on_dissimilar_tasks():
    """That a warning is raised to our users if computing scores across tasks
    with dissimilar names.

    This is allowed as a convenience way of computing scores between e.g.
    'y' and 'my-model::y' without needing to add an explicit renaming step.
    The disadvantage is that it has the risk of silently allowing unrelated
    tasks to be compared for scoring.
    """

    references = datasets.DatasetDict(
        {
            "train": datasets.Dataset.from_dict(
                {
                    "x": [1, 2, 3],
                    "y": [3, 2, 1],
                    "z": [2, 1, 3],
                },
            ),
        },
    )

    predictions = datasets.DatasetDict(
        {
            "train": datasets.Dataset.from_dict(
                {
                    "x": [1, 2, 3],
                    "my-model::y": [3, 2, 1],
                    "another_task": [2, 1, 3],
                },
            ),
        },
    )

    metrics = molflux.metrics.load_suite("regression")

    with pytest.warns(UserWarning, match="dissimilar names: 'z' vs. 'another_task'"):
        compute_scores(predictions=predictions, references=references, metrics=metrics)



"""
Tests for API features advertised as part of the Representation protocol.
"""

from typing import Protocol

from molflux.features.representation import Representation


def test_class_is_protocol():
    """That Representation is a protocol."""
    assert issubclass(type(Representation), type(Protocol))


def test_has_init():
    """That Representation can be initialised."""
    assert hasattr(Representation, "__init__")
    assert callable(Representation.__init__)


def test_has_metadata():
    """That Representation has a metadata property."""
    assert hasattr(Representation, "metadata")
    assert not callable(Representation.metadata)


def test_has_name():
    """That Representation has a name property."""
    assert hasattr(Representation, "name")
    assert not callable(Representation.name)


def test_has_tag():
    """That Representation has a tag property."""
    assert hasattr(Representation, "tag")
    assert not callable(Representation.tag)


def test_has_featurise():
    """That Representation has a featurise method."""
    assert hasattr(Representation, "featurise")
    assert callable(Representation.featurise)


def test_has_reset_state():
    """That Representation has a reset_state method."""
    assert hasattr(Representation, "reset_state")
    assert callable(Representation.reset_state)


def test_has_udpate_state():
    """That Representation has an update_state method."""
    assert hasattr(Representation, "update_state")
    assert callable(Representation.update_state)



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Literal, TypeVar

from pydantic.v1 import dataclasses

import datasets

if TYPE_CHECKING:
    import numpy as np

from molflux.modelzoo.model import ClassificationMixin, ModelBase, ModelConfig
from molflux.modelzoo.typing import Classes, PredictionResult
from molflux.modelzoo.utils import get_concatenated_array, pick_features

try:
    import pyod.models.base
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("pyod", e) from None

ProbabilityConversionMethod = Literal["linear", "unify"]


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class PyODModelConfig(ModelConfig):
    def __post_init_post_parse__(self) -> None:
        if self.y_features:
            if len(self.y_features) != 1:
                raise ValueError(
                    f"Unsupervised learning models only accepts an optional y_features "
                    f"list with a single molflux: got {self.y_features}",
                )
        else:
            # TODO (dm 12 Oct 23) Is this sensible?
            self.train_features = self.x_features
            self.y_features = ["outliers"]


_PyODModelConfigT = TypeVar("_PyODModelConfigT", bound=PyODModelConfig)


class PyODModelBase(ModelBase[_PyODModelConfigT], ABC):
    """Base model for all models based on the PyOD API"""

    @abstractmethod
    def _instantiate_model(self) -> pyod.models.base.BaseDetector: ...

    def _train(self, train_data: datasets.Dataset, **kwargs: Any) -> Any:
        x_data = pick_features(train_data, self.x_features)
        X = get_concatenated_array(x_data, self.x_features)

        # instantiate model
        self.model = self._instantiate_model()

        # train (unsupervised learning)
        self.model.fit(X)

    def _predict(self, data: datasets.Dataset, **kwargs: Any) -> PredictionResult:
        display_names = self._predict_display_names

        if not len(data):
            return {display_name: [] for display_name in display_names}

        x_data = pick_features(data, self.x_features)
        X = get_concatenated_array(x_data, self.x_features)
        y_predict: np.ndarray = self.model.predict(X, return_confidence=False)

        # Outlier detectors are single task models
        return {display_name: y_predict.tolist() for display_name in display_names}


class PyODClassificationMixin(ClassificationMixin):
    model: pyod.models.base.BaseDetector
    tag: str

    @property
    def classes(self) -> Classes:
        if not hasattr(self.model, "_classes"):
            return {}
        return {self.model.y_features[0]: list(self.model._classes)}

    def _predict_proba(
        self,
        data: datasets.Dataset,
        method: ProbabilityConversionMethod = "linear",
        **kwargs: Any,
    ) -> PredictionResult:
        display_names = self._predict_proba_display_names

        if not len(data):
            return {display_name: [] for display_name in display_names}

        x_data = pick_features(data, self.x_features)
        X = get_concatenated_array(x_data, self.x_features)
        y_proba: np.ndarray = self.model.predict_proba(
            X,
            method=method,
            return_confidence=False,
        )
        # Outlier detectors are single task models
        return {display_name: y_proba.tolist() for display_name in display_names}



from typing import Any, Protocol, TypeGuard, Union, runtime_checkable

from molflux.modelzoo.typing import Classes, DataFrameLike, Features, PredictionResult


@runtime_checkable
class Estimator(Protocol):
    """The public protocol for molflux modelzoo Models."""

    def __init__(self, **kwargs: Any):
        """Initialises the model."""

    @property
    def metadata(self) -> dict[str, Any]:
        """Model metadata."""

    @property
    def name(self) -> str:
        """The canonical name of the model."""

    @property
    def tag(self) -> str:
        """The arbitrary tag name of the model."""

    @property
    def config(self) -> dict[str, Any]:
        """The essential config that fully defines the model."""

    @property
    def x_features(self) -> Features:
        """The features that the model has been trained on."""

    @property
    def y_features(self) -> Features:
        """The features that the model predicts."""

    def train(
        self,
        train_data: DataFrameLike | dict[str, DataFrameLike],
        validation_data: DataFrameLike | dict[str, DataFrameLike] | None = None,
        **kwargs: Any,
    ) -> Any:
        """Trains the model."""

    def predict(self, data: DataFrameLike, **kwargs: Any) -> PredictionResult:
        """Performs a model prediction."""

    def as_dir(self, directory: str) -> None:
        """Serialises the backend model objects in a directory."""

    def from_dir(self, directory: str) -> None:
        """Deserialises and sets the backend model objects serialised by `as_dir`."""


@runtime_checkable
class SupportsClassification(Estimator, Protocol):
    @property
    def classes(self) -> Classes:
        """Returns the model's unique class labels"""

    def predict_proba(self, data: DataFrameLike, **kwargs: Any) -> PredictionResult:
        """Returns probability estimates for the data."""


@runtime_checkable
class SupportsCovariance(Estimator, Protocol):
    """The public protocol for Models supporting covariance matrix for multivariate predictions."""

    def predict_with_covariance(
        self,
        data: DataFrameLike,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        """Performs a model prediction with associated covariance matrix."""


@runtime_checkable
class SupportsPredictionInterval(Estimator, Protocol):
    """The public protocol for Models supporting prediction interval."""

    def predict_with_prediction_interval(
        self,
        data: DataFrameLike,
        confidence: float,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        """Performs a model prediction with associated prediction interval."""


@runtime_checkable
class SupportsSampling(Estimator, Protocol):
    """The public protocol for Models supporting sampling."""

    def sample(
        self,
        data: DataFrameLike,
        n_samples: int,
        **kwargs: Any,
    ) -> PredictionResult:
        """Performs a model prediction with associated samples"""


@runtime_checkable
class SupportsStandardDeviation(Estimator, Protocol):
    """The public protocol for Models supporting standard deviation."""

    def predict_with_std(
        self,
        data: DataFrameLike,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        """Performs a model prediction with associated standard deviation."""


@runtime_checkable
class SupportsUncertaintyCalibration(Estimator, Protocol):
    """The public protocol for Models supporting uncertainty calibrations."""

    def calibrate_uncertainty(self, data: DataFrameLike, **kwargs: Any) -> Any:
        """Calibrates the uncertainty of predictions on a validation dataset"""


def supports_classification(model: Any) -> TypeGuard[SupportsClassification]:
    """Returns True if the given model is a classifier."""
    return isinstance(model, SupportsClassification)


def supports_covariance(model: Any) -> TypeGuard[SupportsCovariance]:
    """Return True if the given model supports covariance."""
    return isinstance(model, SupportsCovariance)


def supports_prediction_interval(model: Any) -> TypeGuard[SupportsPredictionInterval]:
    """Return True if the given model supports prediction interval"""
    return isinstance(model, SupportsPredictionInterval)


def supports_sampling(model: Any) -> TypeGuard[SupportsSampling]:
    """Return True if the given model supports sampling."""
    return isinstance(model, SupportsSampling)


def supports_std(model: Any) -> TypeGuard[SupportsStandardDeviation]:
    """Return True if the given model supports standard deviation."""
    return isinstance(model, SupportsStandardDeviation)


def supports_uncertainty_calibration(
    model: Any,
) -> TypeGuard[SupportsUncertaintyCalibration]:
    """Return True if the given model supports uncertainty calibration."""
    return isinstance(model, SupportsUncertaintyCalibration)


Model = Union[
    Estimator,
    SupportsClassification,
    SupportsCovariance,
    SupportsPredictionInterval,
    SupportsSampling,
    SupportsStandardDeviation,
    SupportsUncertaintyCalibration,
]
Models = dict[str, Model]



import logging

import molflux.core
import molflux.datasets
import molflux.features
import molflux.modelzoo
import molflux.splits

from .version import version as __version__

logger = logging.getLogger(__name__)



from collections.abc import Callable
from dataclasses import asdict
from typing import Any, Literal

import numpy as np
import scipy.stats as st
from pydantic.v1 import dataclasses

import datasets
from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import (
    ModelBase,
    ModelConfig,
    PredictionIntervalMixin,
    SamplingMixin,
    StandardDeviationMixin,
    UncertaintyCalibrationMixin,
)
from molflux.modelzoo.typing import PredictionResult
from molflux.modelzoo.utils import validate_features

_DESCRIPTION = """
A core model that predicts the average (median/mean) of the dataset for each predicted variable
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
average_method: Literal, default = "median". Method by which to summarise the average of the dataset.
    Available options are median or mean. median is preferred since it is robust to outliers.
deviation_method: Literal, default = "mad". Method by which to summarise the deviation of the dataset.
    Available options are std or mad. Note that mad is the median absolute deviation which is a version
    of standard deviation robust to outlier values.
"""

# mapping from method strings to fns
_AVERAGE_METHOD_DICT: dict[str, Callable] = {
    "median": np.nanmedian,
    "mean": np.nanmean,
}
_DEVIATION_METHOD_DICT: dict[str, Callable] = {
    "std": np.nanstd,
    "mad": st.median_abs_deviation,
}


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class AverageRegressorConfig(ModelConfig):
    average_method: Literal["median", "mean"] = "median"
    deviation_method: Literal["std", "mad"] = "mad"


class AverageRegressor(
    UncertaintyCalibrationMixin,
    PredictionIntervalMixin,
    StandardDeviationMixin,
    SamplingMixin,
    ModelBase[AverageRegressorConfig],
):
    @property
    def config(self) -> dict[str, Any]:
        return asdict(self.model_config)

    @property
    def _config_builder(self) -> type[AverageRegressorConfig]:
        return AverageRegressorConfig

    def _info(self) -> ModelInfo:
        """Initialises the ModelInfo object.

        To be implemented by subclasses.
        """
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _train(self, train_data: datasets.Dataset, **kwargs: Any) -> Any:
        # validate y features as well
        validate_features(train_data, self.y_features)

        average_fn = _AVERAGE_METHOD_DICT[self.model_config.average_method]
        deviation_fn = _DEVIATION_METHOD_DICT[self.model_config.deviation_method]
        # apply to each variable and save result as the "model"
        self.model = {
            "mu": [average_fn(train_data[y]) for y in self.y_features],
            "sigma": [deviation_fn(train_data[y]) for y in self.y_features],
        }

    def _predict(self, data: datasets.Dataset, **kwargs: Any) -> PredictionResult:
        display_names = self._predict_display_names

        if not len(data):
            return {display_name: [] for display_name in display_names}

        return {
            display_name: np.full(len(data), mu).tolist()
            for display_name, mu in zip(display_names, self.model["mu"], strict=False)
        }

    def _predict_with_std(
        self,
        data: datasets.Dataset,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        (
            prediction_display_names,
            prediction_std_display_names,
        ) = self._predict_with_std_display_names

        if not len(data):
            return {display_name: [] for display_name in prediction_display_names}, {
                display_name: [] for display_name in prediction_std_display_names
            }

        return {
            display_name: np.full(len(data), mu).tolist()
            for display_name, mu in zip(
                prediction_display_names,
                self.model["mu"],
                strict=False,
            )
        }, {
            display_name: np.full(len(data), sigma).tolist()
            for display_name, sigma in zip(
                prediction_std_display_names,
                self.model["sigma"],
                strict=False,
            )
        }

    def _predict_with_prediction_interval(
        self,
        data: datasets.Dataset,
        confidence: float,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        (
            prediction_display_names,
            prediction_interval_display_names,
        ) = self._predict_with_prediction_interval_display_names

        if not len(data):
            return {display_name: [] for display_name in prediction_display_names}, {
                display_name: [] for display_name in prediction_interval_display_names
            }

        # get the result dictionary of means and standard deviations
        prediction_mean_results, prediction_std_results = self._predict_with_std(data)

        # for each column, map the mean and standard deviation to a prediction interval
        prediction_results: PredictionResult = {}
        prediction_prediction_interval_results: PredictionResult = {}
        for prediction_display_name, prediction_interval_display_name, mean, std in zip(
            prediction_display_names,
            prediction_interval_display_names,
            prediction_mean_results.values(),
            prediction_std_results.values(),
            strict=False,
        ):
            # compute the prediction interval
            lower_bound, upper_bound = st.norm.interval(
                confidence,
                loc=mean,
                scale=std,
            )

            # where nans occur (ex. from a 0 standard deviation), use the mean instead
            lower_bound = np.where(~np.isnan(lower_bound), lower_bound, mean)
            upper_bound = np.where(~np.isnan(upper_bound), upper_bound, mean)

            prediction_results[prediction_display_name] = mean
            prediction_prediction_interval_results[prediction_interval_display_name] = (
                list(zip(lower_bound, upper_bound, strict=False))
            )

        return prediction_results, prediction_prediction_interval_results

    def _sample(
        self,
        data: datasets.Dataset,
        n_samples: int,
        **kwargs: Any,
    ) -> PredictionResult:
        display_names = self._sample_display_names

        if not len(data):
            return {display_name: [] for display_name in display_names}

        prediction_mean_results, prediction_std_results = self._predict_with_std(data)

        prediction_results: PredictionResult = {}
        for display_name, means, stds in zip(
            display_names,
            prediction_mean_results.values(),
            prediction_std_results.values(),
            strict=False,
        ):
            samples = np.random.normal(means, stds, (n_samples, len(means))).T
            prediction_results[display_name] = samples.tolist()

        return prediction_results

    def _calibrate_uncertainty(self, data: datasets.Dataset, **kwargs: Any) -> Any:
        pass



from __future__ import annotations

from typing import TYPE_CHECKING, Any

try:
    from molflux.features.representations.rdkit.reaction._drfp_vendored import (
        DrfpEncoder,
    )
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

if TYPE_CHECKING:
    from molflux.features.typing import Fingerprint, SmilesArray

_DESCRIPTION = """
The DRFP algorithm takes a reaction SMILES as an input and creates a binary
fingerprint based on the symmetric difference of two sets containing the
circular molecular n-grams generated from the molecules listed left and right
from the reaction arrow, respectively, without the need for distinguishing
between reactants and reagents.

https://pubs.rsc.org/en/content/articlehtml/2022/dd/d1dd00006c
"""


class DRFP(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: SmilesArray,
        length: int = 2048,
        min_radius: int = 0,
        max_radius: int = 3,
        include_rings: bool = True,
        root_central_atom: bool = True,
        include_hydrogens: bool = False,
        **kwargs: Any,
    ) -> dict[str, list[Fingerprint]]:
        """Featurises the input molecules as DRFP (differential
        reaction fingerprint.

        Args:
            samples: The array of samples to featurise.
            length: The folded length of the fingerprint (the parameter for the modulo hashing)
            min_radius: The minimum radius of a substructure (0 includes single atoms)
            max_radius: The maximum radius of a substructure
            include_rings: Whether to include full rings as substructures
            atom_index_mapping: Return the atom indices of mapped substructures for each reaction
            root_central_atom: Whether to root the central atom of substructures when generating SMILES
            include_hydrogens: Whether to explicitly include hydrogens in the molecular graph

        Returns:
            DRFP fingerpints.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation("drfp")
            >>> samples = ["CO.O.O=C(NC(=S)Nc1nc(-c2ccccc2)cs1)c1ccccc1.[Na+].[OH-]>>NC(=S)Nc1nc(-c2ccccc2)cs1"]
            >>> representation.featurise(samples, length=4)
            {'drfp': [[1, 1, 1, 1]]}
        """
        del kwargs
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]

        fingerprints: list[Fingerprint] = []
        for sample in samples:
            with featurisation_error_harness(sample):
                fingerprint = DrfpEncoder.encode_one(
                    sample,
                    length=length,
                    min_radius=min_radius,
                    max_radius=max_radius,
                    include_rings=include_rings,
                    root_central_atom=root_central_atom,
                    include_hydrogens=include_hydrogens,
                )
                fingerprints.append(fingerprint)

        return {self.tag: fingerprints}



import pytest

import molflux.metrics.catalogue
from molflux.metrics.catalogue import get_metric_cls, list_metrics, register_metric

catalogue_obj = molflux.metrics.catalogue.METRICS_CATALOGUE


def test_catalogue_is_filled():
    """That the catalogue has been filled."""
    catalogue = list_metrics()
    assert len(catalogue)


def test_list_metrics_returns_sorted_names():
    """That entries in the catalogue are returned sorted by name."""
    catalogue = list_metrics()
    assert sorted(catalogue.keys()) == list(catalogue.keys())
    for metrics in catalogue.values():
        assert sorted(metrics) == metrics


def test_list_metrics_lists_names_of_metrics_in_catalogue():
    """That all metrics existing in the catalogue are returned by list_metrics."""
    view = list_metrics()
    names_in_view = [name for names in view.values() for name in names]
    assert sorted(names_in_view) == sorted(catalogue_obj.keys())


def test_register_metric(monkeypatch):
    """That can register a new metric in the catalogue."""
    # set up a mock empty catalogue to leave real one untouched
    monkeypatch.setattr(molflux.metrics.catalogue, "METRICS_CATALOGUE", {})

    new_metric_kind = "testing"
    new_metric_name = "pytest_metric"
    assert new_metric_name not in list_metrics()

    @register_metric(kind=new_metric_kind, name=new_metric_name)
    class PytestMetric: ...

    new_catalogue = list_metrics()
    assert new_metric_kind in new_catalogue
    assert new_metric_name in new_catalogue[new_metric_kind]


def test_get_metric_not_in_catalogue_raises_not_implemented_error():
    """That getting a non-existent metric raises a NotImplementedError."""
    metric_name = "non-existent-metric"
    with pytest.raises(
        NotImplementedError,
        match=f"Metric {metric_name!r} is not available",
    ):
        get_metric_cls(metric_name)



from typing import Any

try:
    from openeye.oemolprop import OEGetHBondDonorCount
    from openeye.oequacpac import OESetNeutralpHModel
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("openeye", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.openeye._utils import to_digit, to_oemol
from molflux.features.typing import MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
The number of hydrogen-bond donors in a molecule based on the definition in the
work of Mills and Dean ([MillsDean-1996]) and also in the book by
Jeffrey ([Jeffrey-1997]).

It is defined to be the number of hydrogen atoms on nitrogen, oxygen, or
sulfur atoms.
"""


class NumDonors(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        digitise: bool = False,
        start: int = 0,
        stop: int = 5,
        num: int = 5,
        set_neutral_ph: bool = False,
        **kwargs: Any,
    ) -> dict[str, list[int]]:
        """Calculates the number of hydrogen-bond donors for each input molecule.

        Args:
            samples: The sample molecules to featurise.
            digitise: Whether to digitise the results. Defaults to `False`.
            start: If `digitise=True`, the start of the binning range. Defaults
                to 0.
            stop: If `digitise=True`, the end of the binning range. Defaults
                to 0.
            num: If `digitise=True`, the number of bins. Defaults to 5.
            set_neutral_ph: Whether to attempt to set mol to the most
                energetically favorable ionization state for pH=7.4 using a
                rule-based system. Multiple states are not enumerated for any
                molecule. In cases where the ionization is near 7.4 and could be
                represented by a mixture of states, the state that is likely to
                be most populated is chosen.

        Returns:
            The number of hydrogen-bond donors of each input molecule.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('num_donors')
            >>> samples = ['c1ccccc1', "NH"]
            >>> representation.featurise(samples=samples)
            {'num_donors': [0, 3]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        results = []
        for sample in samples:
            with featurisation_error_harness(sample):
                mol = to_oemol(sample)
                if set_neutral_ph:
                    new = mol.CreateCopy()
                    OESetNeutralpHModel(new)
                    result = OEGetHBondDonorCount(mol=new)
                else:
                    result = OEGetHBondDonorCount(mol=mol)

                if digitise:
                    result = to_digit(result, start=start, stop=stop, num=num).tolist()

                results.append(result)

        return {self.tag: results}



from collections.abc import Iterator
from typing import Any

import numpy as np
from sklearn.utils import check_random_state

from molflux.splits.bases import SplittingStrategyBase
from molflux.splits.info import SplittingStrategyInfo
from molflux.splits.typing import ArrayLike, SplitIndices, Splittable
from molflux.splits.utils import partition

_DESCRIPTION = """
Random permutation cross-validator.
"""


class ShuffleSplit(SplittingStrategyBase):
    def _info(self) -> SplittingStrategyInfo:
        return SplittingStrategyInfo(
            description=_DESCRIPTION,
        )

    def _split(
        self,
        dataset: Splittable,
        y: ArrayLike | None = None,
        groups: ArrayLike | None = None,
        *,
        n_splits: int = 1,
        train_fraction: float = 0.8,
        validation_fraction: float = 0.1,
        test_fraction: float = 0.1,
        seed: int | None = None,
        **kwargs: Any,
    ) -> Iterator[SplitIndices]:
        """
        Args:
            dataset: The data to be split.
            y (optional): The target variable for supervised learning problems.
            groups (optional): Group labels for the samples used while splitting the dataset.
            n_splits (optional): The number of splits to generate. Defaults to 1.
            train_fraction (optional): The proportion of the dataset to include in the train split.
                Defaults to 0.8.
            validation_fraction: The proportion of the dataset to include in the validation split.
                Defaults to 0.1.
            test_fraction: The proportion of the dataset to include in the test split.
                Defaults to 0.1.
            seed (optional): Controls the shuffling applied to the data before applying
                the split. Pass an int for reproducible output across multiple function calls.
                Defaults to None.
        """
        np.testing.assert_almost_equal(
            train_fraction + validation_fraction + test_fraction,
            1.0,
        )

        train_cutoff, validation_cutoff = partition(
            dataset,
            train_fraction,
            validation_fraction,
        )
        indices = range(len(dataset))

        rng = check_random_state(seed)
        for _ in range(n_splits):
            permutation = rng.permutation(indices)
            train_indices, validation_indices, test_indices = np.split(
                permutation,
                [train_cutoff, validation_cutoff],
            )
            yield train_indices, validation_indices, test_indices



from typing import Any

try:
    from openeye.oechem import OENetCharge
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("openeye", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.openeye._utils import to_digit, to_oemol
from molflux.features.typing import MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
Determines the net charge on a molecule.

If the molecule has specified partial charges, this featuriser returns the sum
of the partial charges rounded to an integer. Otherwise this featuriser returns
the sum of the formal charges on each atom of the molecule.
"""


class NetCharge(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        digitise: bool = False,
        start: int = -2,
        stop: int = 2,
        num: int = 4,
        **kwargs: Any,
    ) -> dict[str, list[int]]:
        """Calculates the net charge of each input molecule.

        Args:
            samples: The sample molecules to featurise.
            digitise: Whether to digitise the results. Defaults to `False`.
            start: If `digitise=True`, the start of the binning range. Defaults
                to -2.
            stop: If `digitise=True`, the end of the binning range. Defaults
                to 2.
            num: If `digitise=True`, the number of bins. Defaults to 4.

        Returns:
            The net charge of each input molecule.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('net_charge')
            >>> samples = ['c1ccccc1', '[Na+]Cl']
            >>> representation.featurise(samples)
            {'net_charge': [0, 1]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        results = []
        for sample in samples:
            with featurisation_error_harness(sample):
                mol = to_oemol(sample)
                result = OENetCharge(mol)

                if digitise:
                    result = to_digit(result, start=start, stop=stop, num=num).tolist()

                results.append(result)

        return {self.tag: results}



from collections.abc import Callable, Sized
from decimal import Decimal
from types import FunctionType


def copyfunc(func: Callable) -> FunctionType:
    """Makes a complete copy of a given function.

    References:
        https://github.com/huggingface/datasets/blob/95193ae61e92aa537d0c65d37a1fd9d2393aae89/src/datasets/utils/py_utils.py#L681
    """
    result = FunctionType(
        func.__code__,
        func.__globals__,
        func.__name__,
        func.__defaults__,
        func.__closure__,
    )
    result.__kwdefaults__ = func.__kwdefaults__
    return result


def partition(dataset: Sized, *fractions: float) -> tuple[int, ...]:
    """Returns the indices where to split the dataset to partition it according
    to the requested fractions.

    The output can then be passed to `np.split` to obtain n+1 sub-arrays.

    Examples:
        train_cutoff, validation_cutoff = partition(dataset, train_fraction, validation_fraction)
        train_split, validation_split, test_split = np.split(dataset, [train_cutoff, validation_cutoff])

    References:
        https://docs.python.org/3/tutorial/floatingpoint.html
        https://docs.python.org/3/library/decimal.html
    """
    n_samples = len(dataset)

    cutoffs = []
    partition_fraction = Decimal(0)
    for fraction in fractions:
        # Obtain an exact floating point sum
        partition_fraction = Decimal(str(partition_fraction)) + Decimal(str(fraction))
        cutoff = int(partition_fraction * n_samples)

        cutoffs.append(cutoff)

    return tuple(cutoffs)



from collections.abc import Mapping, Sequence
from os import PathLike as OSPathLike

from molflux.datasets.typing import DataFiles, HFDataFiles


def _is_cloud_file(file: str) -> bool:
    """Validates if file is a cloud file"""
    if file.startswith("s3://"):
        return True
    return False


def is_cloud_data(data_files: HFDataFiles | None) -> bool:
    """Validates if data files refer to data in the cloud."""

    if data_files is None:
        return False

    if isinstance(data_files, str):
        return _is_cloud_file(data_files)

    if isinstance(data_files, Sequence):
        return _is_cloud_file(data_files[0])

    if isinstance(data_files, Mapping):
        files = list(data_files.values())
        file = files[0]

        if isinstance(file, str):
            return _is_cloud_file(file)

        if isinstance(file, Sequence):
            return _is_cloud_file(file[0])

    raise ValueError("Could not resolve if data files are cloud data files")


def sanitise_data_files_as_str(data_files: DataFiles) -> HFDataFiles:
    """Coerces Path-like DataFiles to str-like DataFiles.

    This is to simplify and sanitise the inputs to match the huggingface
    datasets DataFiles API.
    """

    if isinstance(data_files, (str, OSPathLike)):
        return str(data_files)

    if isinstance(data_files, Sequence):
        return [str(data_file) for data_file in data_files]

    if isinstance(data_files, Mapping):
        return {k: sanitise_data_files_as_str(files) for k, files in data_files.items()}  # type: ignore[misc]

    raise TypeError(f"Unsupported data files type: {type(data_files)}")



from typing import Any

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.typing import ArrayLike

_DESCRIPTION = """
A generic representation that sums iterable elements.
"""


class Sum(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(self, *columns: ArrayLike, **kwargs: Any) -> dict[str, list[Any]]:
        """Sums every input column.

        Args:
            *columns: The data to sum. Should consist of iterable elements.

        Returns:
            A dictionary of the sum of columns.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('sum')
            >>> columns = [1, 2, 3, 4], [10, 20, 30, 40], [100, 200, 300, 400]
            >>> representation.featurise(*columns)
            {'sum': [111, 222, 333, 444]}
        """

        return {f"{self.tag}": [sum(t) for t in zip(*columns, strict=False)]}



import pytest
from cloudpathlib import AnyPath

import datasets
import molflux.modelzoo
from molflux.core import load_model, save_model

parametrized_output_directory_fixtures = [
    "tmp_path",
    "fixture_test_bucket",
    "fixture_nested_tmp_path",
    "fixture_nested_test_bucket",
]


@pytest.mark.parametrize(
    "output_directory_fixture",
    parametrized_output_directory_fixtures,
)
def test_save_model_creates_standardised_artefact(output_directory_fixture, request):
    """That models are saved into a standardised artefact directory."""
    output_directory = request.getfixturevalue(output_directory_fixture)

    dataset = datasets.Dataset.from_dict({"x": [1, 2, 3], "y": [1, 2, 3]})
    featurisation_metadata = {"version": 1, "config": []}
    model = molflux.modelzoo.load_model(
        "linear_regressor",
        x_features=["x"],
        y_features=["y"],
    )

    model.train(dataset)

    artefact_directory = save_model(
        model,
        path=output_directory,
        featurisation_metadata=featurisation_metadata,
    )
    artefact_path = AnyPath(artefact_directory)

    assert artefact_path.is_dir()  # type: ignore[attr-defined]

    ls = list(artefact_path.glob("**/*"))  # type: ignore[attr-defined]

    assert artefact_path / "featurisation_metadata.json" in ls  # type: ignore[operator]

    # optional for now (no clients depend on this)
    assert artefact_path / "requirements.txt" in ls  # type: ignore[operator]

    # these are an implementation detail of molflux.modelzoo.save_model()
    # we test them here just in case but could technically change without worries
    # as long as molflux.modelzoo also knows how to read them back
    assert artefact_path / "model_config.json" in ls  # type: ignore[operator]
    assert artefact_path / "model_artefacts" in ls  # type: ignore[operator]


def test_save_model_with_no_featurisation_metadata_raises_warning(tmp_path):
    """That a warning is raised if saving models without the associated
    featurisation metadata that was used to featurise the training dataset.

    Users that do not want to associate featurisation metadata should be able
    to communicate that by passing in None or an empty dict.
    """

    dataset = datasets.Dataset.from_dict({"x": [1, 2, 3], "y": [1, 2, 3]})
    model = molflux.modelzoo.load_model(
        "linear_regressor",
        x_features=["x"],
        y_features=["y"],
    )

    model.train(dataset)

    with pytest.warns(UserWarning, match="featurisation metadata"):
        save_model(model, path=tmp_path, featurisation_metadata=None)

    with pytest.warns(UserWarning, match="featurisation metadata"):
        save_model(model, path=tmp_path, featurisation_metadata={})


@pytest.mark.parametrize(
    "output_directory_fixture",
    parametrized_output_directory_fixtures,
)
def test_can_load_back_saved_model(output_directory_fixture, request):
    """That can load a model that has been saved."""
    output_directory = request.getfixturevalue(output_directory_fixture)

    dataset = datasets.Dataset.from_dict({"x": [1, 2, 3], "y": [1, 2, 3]})
    featurisation_metadata = {"version": 1, "config": []}
    model = molflux.modelzoo.load_model(
        "linear_regressor",
        x_features=["x"],
        y_features=["y"],
    )

    model.train(dataset)
    model_path = save_model(
        model,
        path=output_directory,
        featurisation_metadata=featurisation_metadata,
    )

    reloaded_model = load_model(model_path)

    assert reloaded_model.name == "linear_regressor"



from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.sklearn import SKLearnModelBase

try:
    from sklearn.linear_model import LinearRegression as SKLinearRegression
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from None


_DESCRIPTION = """
This is an sklearn linear regression model.

LinearRegression fits a linear model with coefficients w = (w1, ..., wp)
to minimize the residual sum of squares between the observed targets in
the dataset, and the targets predicted by the linear approximation.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
fit_intercept : bool, default=True
    Whether to calculate the intercept for this model. If set
    to False, no intercept will be used in calculations
    (i.e. data is expected to be centered).
copy_X : bool, default=True
    If True, X will be copied; else, it may be overwritten.
n_jobs : int, default=None
    The number of jobs to use for the computation. This will only provide
    speedup in case of sufficiently large problems, that is if firstly
    `n_targets > 1` and secondly `X` is sparse or if `positive` is set
    to `True`. ``None`` means 1 unless in a
    :obj:`joblib.parallel_backend` context. ``-1`` means using all
    processors.
positive : bool, default=False
    When set to ``True``, forces the coefficients to be positive. This
    option is only supported for dense arrays.
"""


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class LinearRegressionConfig(ModelConfig):
    fit_intercept: bool = True
    copy_X: bool = True
    n_jobs: int | None = None
    positive: bool = False


class LinearRegressor(SKLearnModelBase[LinearRegressionConfig]):
    @property
    def _config_builder(self) -> type[LinearRegressionConfig]:
        return LinearRegressionConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> SKLinearRegression:
        config = self.model_config
        return SKLinearRegression(
            fit_intercept=config.fit_intercept,
            copy_X=config.copy_X,
            n_jobs=config.n_jobs,
            positive=config.positive,
        )



if __name__ == "__main__":
    import h5py

    import datasets

    dl_manager = datasets.DownloadManager()

    h5_file_path = dl_manager.download_and_extract(
        "https://figshare.com/ndownloader/files/18112775",
    )
    h5_file = h5py.File(h5_file_path, "r")

    new_h5_file = h5py.File("data.h5", "w")

    for key in list(h5_file.keys())[:5]:
        new_h5_file.create_group(key)
        new_h5_file[key].create_dataset(
            "atomic_numbers",
            data=h5_file[key]["atomic_numbers"],
        )
        new_h5_file[key].create_dataset(
            "ccsd(t)_cbs.energy",
            data=h5_file[key]["ccsd(t)_cbs.energy"][:2],
        )
        new_h5_file[key].create_dataset(
            "coordinates",
            data=h5_file[key]["coordinates"][:2],
        )
        new_h5_file[key].create_dataset(
            "hf_dz.energy",
            data=h5_file[key]["hf_dz.energy"][:2],
        )
        new_h5_file[key].create_dataset(
            "hf_qz.energy",
            data=h5_file[key]["hf_qz.energy"][:2],
        )
        new_h5_file[key].create_dataset(
            "hf_tz.energy",
            data=h5_file[key]["hf_tz.energy"][:2],
        )
        new_h5_file[key].create_dataset(
            "mp2_dz.corr_energy",
            data=h5_file[key]["mp2_dz.corr_energy"][:2],
        )
        new_h5_file[key].create_dataset(
            "mp2_qz.corr_energy",
            data=h5_file[key]["mp2_qz.corr_energy"][:2],
        )
        new_h5_file[key].create_dataset(
            "mp2_tz.corr_energy",
            data=h5_file[key]["mp2_tz.corr_energy"][:2],
        )
        new_h5_file[key].create_dataset(
            "npno_ccsd(t)_dz.corr_energy",
            data=h5_file[key]["npno_ccsd(t)_dz.corr_energy"][:2],
        )
        new_h5_file[key].create_dataset(
            "npno_ccsd(t)_tz.corr_energy",
            data=h5_file[key]["npno_ccsd(t)_tz.corr_energy"][:2],
        )
        new_h5_file[key].create_dataset(
            "tpno_ccsd(t)_dz.corr_energy",
            data=h5_file[key]["tpno_ccsd(t)_dz.corr_energy"][:2],
        )
        new_h5_file[key].create_dataset(
            "wb97x_dz.cm5_charges",
            data=h5_file[key]["wb97x_dz.cm5_charges"][:2],
        )
        new_h5_file[key].create_dataset(
            "wb97x_dz.dipole",
            data=h5_file[key]["wb97x_dz.dipole"][:2],
        )
        new_h5_file[key].create_dataset(
            "wb97x_dz.energy",
            data=h5_file[key]["wb97x_dz.energy"][:2],
        )
        new_h5_file[key].create_dataset(
            "wb97x_dz.forces",
            data=h5_file[key]["wb97x_dz.forces"][:2],
        )
        new_h5_file[key].create_dataset(
            "wb97x_dz.hirshfeld_charges",
            data=h5_file[key]["wb97x_dz.hirshfeld_charges"][:2],
        )
        new_h5_file[key].create_dataset(
            "wb97x_dz.quadrupole",
            data=h5_file[key]["wb97x_dz.quadrupole"][:2],
        )
        new_h5_file[key].create_dataset(
            "wb97x_tz.dipole",
            data=h5_file[key]["wb97x_tz.dipole"][:2],
        )
        new_h5_file[key].create_dataset(
            "wb97x_tz.energy",
            data=h5_file[key]["wb97x_tz.energy"][:2],
        )
        new_h5_file[key].create_dataset(
            "wb97x_tz.forces",
            data=h5_file[key]["wb97x_tz.forces"][:2],
        )
        new_h5_file[key].create_dataset(
            "wb97x_tz.mbis_charges",
            data=h5_file[key]["wb97x_tz.mbis_charges"][:2],
        )
        new_h5_file[key].create_dataset(
            "wb97x_tz.mbis_dipoles",
            data=h5_file[key]["wb97x_tz.mbis_dipoles"][:2],
        )
        new_h5_file[key].create_dataset(
            "wb97x_tz.mbis_octupoles",
            data=h5_file[key]["wb97x_tz.mbis_octupoles"][:2],
        )
        new_h5_file[key].create_dataset(
            "wb97x_tz.mbis_quadrupoles",
            data=h5_file[key]["wb97x_tz.mbis_quadrupoles"][:2],
        )
        new_h5_file[key].create_dataset(
            "wb97x_tz.mbis_volumes",
            data=h5_file[key]["wb97x_tz.mbis_volumes"][:2],
        )

    new_h5_file.close()



import logging
from collections.abc import Iterator
from typing import Any

try:
    import numpy as np
    from rdkit import Chem
    from rdkit.Chem.Scaffolds.MurckoScaffold import MurckoScaffoldSmiles
except ImportError as e:
    from molflux.splits.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from e

from molflux.splits.bases import SplittingStrategyBase
from molflux.splits.info import SplittingStrategyInfo
from molflux.splits.typing import ArrayLike, SplitIndices, Splittable
from molflux.splits.utils import partition

logger = logging.getLogger(__name__)

_DESCRIPTION = """
Class for doing data splits based on the scaffold of small molecules using RDKit.

Group  molecules  based on  the Bemis-Murcko scaffold representation, which identifies rings,
linkers, frameworks (combinations between linkers and rings) and atomic properties  such as
atom type, hibridization and bond order in a dataset of molecules. Then split the groups by
the number of molecules in each group in decreasing order.

IMPORTANT NOTE: If no core can be extracted, a warning will be raised and the SMILES itself is used as a scaffold.

References:
    .. [1] Bemis, Guy W., and Mark A. Murcko. "The properties of known drugs.
    1. Molecular frameworks." Journal of medicinal chemistry 39.15 (1996): 2887-2893.
"""


class Scaffold(SplittingStrategyBase):
    def _info(self) -> SplittingStrategyInfo:
        return SplittingStrategyInfo(
            description=_DESCRIPTION,
        )

    def _split(
        self,
        dataset: Splittable,
        y: ArrayLike | None = None,
        groups: ArrayLike | None = None,
        train_fraction: float = 0.8,
        validation_fraction: float = 0.1,
        test_fraction: float = 0.1,
        include_chirality: bool = False,
        **kwargs: Any,
    ) -> Iterator[SplitIndices]:
        """
        Args:
            y: List of smiles to be used for scaffold split
            groups (optional): Group labels for the samples used while splitting the dataset.
            train_fraction: The proportion of the dataset to include in the train split.
            validation_fraction: The proportion of the dataset to include in the validation split.
            test_fraction: The proportion of the dataset to include in the test split.
            include_chirality (optional): Whether to include chirality in scaffolds or not.
                Defaults to False.

        Examples:
            >>> from molflux.splits import load_splitting_strategy
            >>> strategy = load_splitting_strategy('scaffold_rdkit')
            >>> dataset = ['CCCC', 'CC', 'c1ccncc1']
            >>> folds = strategy.split(dataset=dataset, y=dataset)
        """
        if y is None:
            raise ValueError("""y parameter should be provided for scaffold splits.""")

        np.testing.assert_almost_equal(
            train_fraction + validation_fraction + test_fraction,
            1.0,
        )

        train_cutoff, validation_cutoff = partition(
            dataset,
            train_fraction,
            validation_fraction,
        )

        scaffold_sets = generate_scaffolds(y=y, include_chirality=include_chirality)

        train_indices: list[int] = []
        validation_indices: list[int] = []
        test_indices: list[int] = []

        for scaffold_set in scaffold_sets:
            if len(train_indices) + len(scaffold_set) > train_cutoff:
                if (
                    len(train_indices) + len(validation_indices) + len(scaffold_set)
                    > validation_cutoff
                ):
                    test_indices += scaffold_set
                else:
                    validation_indices += scaffold_set
            else:
                train_indices += scaffold_set

        yield train_indices, validation_indices, test_indices


def _generate_scaffold(smiles: str, include_chirality: bool) -> str:
    """Compute the Bemis-Murcko scaffold for a SMILES string.

    Bemis-Murcko scaffolds are described in DOI: 10.1021/jm9602928.
    They are essentially that part of the molecule consisting of
    rings and the linker atoms between them.

    Args:
        smiles: The SMILES string.
        include_chirality: Whether to include chirality in scaffolds or not.

    Returns:
        The MurckScaffold SMILES from the original SMILES

    References:
        .. [1] Bemis, Guy W., and Mark A. Murcko. "The properties of known drugs.
        1. Molecular frameworks." Journal of medicinal chemistry 39.15 (1996): 2887-2893.
    """

    try:
        mol = Chem.MolFromSmiles(smiles)
        scaffold: str = MurckoScaffoldSmiles(
            mol=mol,
            includeChirality=include_chirality,
        )
        return scaffold
    except Exception:
        logger.warning(
            f"Could not find scaffold for molecule {smiles}. Will use the smiles itself as a scaffold.",
        )
        return smiles


def generate_scaffolds(y: ArrayLike, include_chirality: bool) -> list[list[int]]:
    """Returns all scaffolds from the SMILES string provided.

    Returns:
        List of indices of each scaffold in y.
    """

    scaffolds = {}
    for ind, smiles in enumerate(y):
        scaffold = _generate_scaffold(
            smiles=smiles,
            include_chirality=include_chirality,
        )

        if scaffold not in scaffolds:
            scaffolds[scaffold] = [ind]
        else:
            scaffolds[scaffold].append(ind)

    # Sort from largest to smallest scaffold sets
    scaffolds = {key: sorted(value) for key, value in scaffolds.items()}
    scaffold_sets = [
        scaffold_set
        for (scaffold, scaffold_set) in sorted(
            scaffolds.items(),
            key=lambda x: (len(x[1]), x[1][0]),
            reverse=True,
        )
    ]
    return scaffold_sets



from pydantic.v1 import dataclasses

from molflux.modelzoo.models.lightning.config import ConfigDict, LightningConfig


@dataclasses.dataclass(config=ConfigDict)
class MLPConfig(LightningConfig):
    input_dim: int = 1
    num_layers: int = 2
    hidden_dim: int = 128
    num_tasks: int = 1

    def __post_init__(self) -> None:
        assert self.input_dim is not None, AttributeError("Must specify `input_dim`")

        if self.num_layers == 1:
            assert self.hidden_dim is None, AttributeError(
                "For 1 layer MLP, `hidden_dim` must be set to `None`.",
            )

        self.num_tasks = len(self.y_features)



import pandas as pd
import pytest

from molflux.modelzoo.models.ensemble._combo.utils import (
    check_parameter,
    get_split_indices,
    list_diff,
)

_X_FEATURES = ["X_col_1", "X_col_2"]
_Y_FEATURES = ["y_col"]

train_df = pd.DataFrame(
    [
        [1, 2],
        [1, 3],
        [2, 4],
        [2, 4],
        [2, 5],
    ],
    columns=_X_FEATURES,
)


def test_list_diff_example():
    """test that list diff works"""
    actual = list_diff([1, 2, -9, -3], [1, -3])
    expected = [2, -9]
    assert len(actual) == len(expected)
    assert all(a == b for a, b in zip(actual, expected, strict=False))


def test_list_diff_duplicates():
    """test that list diff removes duplicates properly"""
    actual = list_diff([1, 2, 2], [1, 2])
    expected: list[int] = []
    assert len(actual) == len(expected)
    assert all(a == b for a, b in zip(actual, expected, strict=False))


def test_get_split_indices():
    """test that utility function gets split indices with right number of folds and containing all the indices"""
    train_dataset = train_df
    n_folds = 4
    actual = get_split_indices(train_dataset, n_folds=n_folds)
    flat_actual_list = [item for sublist in actual for item in sublist]
    expected = list(range(train_dataset.shape[0]))
    assert len(actual) == n_folds
    assert len(flat_actual_list) == train_dataset.shape[0]
    assert all(a == b for a, b in zip(flat_actual_list, expected, strict=False))


def test_check_parameter():
    actual = check_parameter(param=5.1, low=4, high=8)
    expected = True
    assert actual == expected
    with pytest.raises(ValueError):
        check_parameter(param=-12, low=0, high=20, param_name="alpha")



from typing import Any

try:
    from openeye.oemolprop import OEGet2dPSA
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("openeye", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.openeye._utils import to_digit, to_oemol
from molflux.features.typing import MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
The Topological polar-surface area (TPSA) for a given molecule.

Topological polar-surface area (TPSA) is based on the algorithm developed by
Ertl et al [Ertl-2000]. In Ertl`s publication, use of TPSA both with and without
accounting for phosphorus and sulfur surface-area is reported. However,
evidence shows that in most PK applications one is better off not counting the
contributions of phosphorus and sulfur atoms toward the total TPSA for a
molecule. This implementation of TPSA allows either inclusion or exclusion of
phosphorus and sulfur surface area.
"""


class TPSA(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        digitise: bool = False,
        start: int = 0,
        stop: int = 140,
        num: int = 14,
        atom_psa: float | None = None,
        s_and_p: bool = False,
        **kwargs: Any,
    ) -> dict[str, list[float]]:
        """Calculates the Topological polar-surface area (TPSA) for each input
        molecule.

        .. warning::

            TPSA values are mildly sensitive to the protonation state of a molecule.

        Args:
            samples: The sample molecules to featurise.
            digitise: Whether to digitise the results. Defaults to `False`.
            start: If `digitise=True`, the start of the binning range. Defaults
                to 0.
            stop: If `digitise=True`, the end of the binning range. Defaults
                to 140.
            num: If `digitise=True`, the number of bins. Defaults to 14.
            atom_psa: Can be used to retrieve the contribution of each atom to
                the total polar surface area.
            s_and_p: Whether sulfur and phosphorus should be counted towards
                the total surface area. Defaults to `False`.

        Returns:
            The Topological polar-surface area (TPSA) for each input molecule.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('tpsa')
            >>> samples = ['C1=CC=C(C=C1)S(=O)O', 'C']
            >>> representation.featurise(samples=samples)
            {'tpsa': [37.29999923706055, 0.0]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        results = []
        for sample in samples:
            with featurisation_error_harness(sample):
                mol = to_oemol(sample)
                result = OEGet2dPSA(mol, atomPSA=atom_psa, SandP=s_and_p)

                if digitise:
                    result = to_digit(result, start=start, stop=stop, num=num).tolist()

                results.append(result)

        return {self.tag: results}



from molflux.metrics.catalogue import fill_catalogue, list_metrics, register_metric
from molflux.metrics.load import (
    load_from_dict,
    load_from_dicts,
    load_from_yaml,
    load_metric,
    load_metrics,
    load_suite,
)
from molflux.metrics.metric import Metric, Metrics
from molflux.metrics.protocols import supports_prediction_intervals
from molflux.metrics.suites.catalogue import list_suites
from molflux.metrics.suites.discovery import import_suites

# Register all plugins at package import time (to fill catalogue)
fill_catalogue()

# import all suites at package import time (to fill catalogue)
import_suites()



"""
Abstract Base Classes for classes implementing the Representation protocol.
"""

import inspect
import logging
import time
import types
from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Any

from molflux import __version__
from molflux.features.info import RepresentationInfo
from molflux.features.naming import camelcase_to_snakecase
from molflux.features.typing import ArrayLike, RepresentationResult
from molflux.features.utils import copyfunc

logger = logging.getLogger(__name__)


class RepresentationBase(ABC):
    """The abstract base class for all concrete representations."""

    def __init__(self, *, tag: str | None = None, **kwargs: Any) -> None:
        """Initialises the representation."""

        # build info
        info = self._info()
        info.name = camelcase_to_snakecase(type(self).__name__)
        info.tag = tag or info.name
        info.version = __version__
        info.featurise_description = (
            info.featurise_description or self._featurise.__doc__ or ""
        )

        # persist info
        self._representation_info = info

        # Update 'featurise' docstring with detailed kwargs description from .info()
        # need to copy method to avoid changing the docstring of every instance
        self.featurise = types.MethodType(copyfunc(self.featurise), self)  # type: ignore[method-assign]
        self.featurise.__func__.__doc__ += (  # type: ignore[attr-defined,operator,unused-ignore]
            self._representation_info.featurise_description
        )

        # The featurisation signature
        self._signature = inspect.signature(self._featurise)

        # Initialise a default state
        self._default_state: dict[str, Any] = {
            parameter.name: parameter.default
            for parameter in self._signature.parameters.values()
            # Positional included in case someone misses `, *, ` in their representation signature.
            if (
                parameter.kind
                in {parameter.POSITIONAL_OR_KEYWORD, parameter.KEYWORD_ONLY}
            )
            # Empty parameters not included (i.e. `samples`, `kwargs`).
            and (parameter.default is not parameter.empty)
        }
        self._state: dict[str, Any] = self._default_state

    def __str__(self) -> str:
        return (
            f"Representation(\n"
            f'\tname: "{self.name}",\n'
            f'\ttag: "{self.tag}",\n'
            f"\tsignature: self.featurise{self._signature},\n"
            f'\tdescription: """{self._representation_info.description}""",\n'
            f'\tusage: """{self._representation_info.featurise_description}"""\n'
            f"\tstate: {self.state!r}\n"
            ")"
        )

    @abstractmethod
    def _info(self) -> RepresentationInfo:
        """Initialises the RepresentationInfo object.

        To be implemented by subclasses.
        """

    @property
    def metadata(self) -> dict[str, Any]:
        return self._representation_info.to_dict()

    @property
    def name(self) -> str:
        return self._representation_info.name

    @property
    def state(self) -> dict[str, Any]:
        return self._state

    @property
    def tag(self) -> str:
        return self._representation_info.tag

    def featurise(self, *columns: ArrayLike, **kwargs: Any) -> RepresentationResult:
        """Featurises the input samples."""

        # Merge explicit keyword arguments with those stored in the state
        kwargs = {**self.state, **kwargs}

        # Safeguard against invalid kwargs
        if not all(k in self._signature.parameters for k in kwargs.keys()):
            unknown_kwargs = [
                k for k in kwargs.keys() if k not in self._signature.parameters
            ]
            raise ValueError(
                f"Unknown featurisation parameter(s): {unknown_kwargs}\n\n"
                f"Expected signature self.featurise{self._signature}",
            )

        # Safeguard against non ArrayLike inputs
        columns = tuple(
            [column]
            if isinstance(column, (str, bytes)) or not isinstance(column, Iterable)
            else column
            for column in columns
        )

        logging_context: dict[str, Any] = {"molflux_representation_tag": self.tag}
        logger.info(
            f"Attempting to featurise samples... {logging_context}",
            extra=logging_context,
        )

        start = time.perf_counter()
        results = self._featurise(*columns, **kwargs)
        elapsed = time.perf_counter() - start

        features = list(results.keys())
        results_shape = [len(results[features[0]]), len(features)]
        try:
            rate = elapsed / (results_shape[0] * results_shape[1])
        except ZeroDivisionError:
            rate = 0

        logging_context.update(
            {
                **logging_context,
                "molflux_representation_featurisation_results_shape": results_shape,
                "molflux_representation_featurisation_time_sec": elapsed,
                "molflux_representation_featurisation_rate_sec_per_feature_per_sample": rate,
            },
        )
        logger.info(f"Featurisation completed {logging_context}", extra=logging_context)

        return results

    @abstractmethod
    def _featurise(self, *columns: ArrayLike, **kwargs: Any) -> RepresentationResult:
        """The featurisation callable to be implemented by subclasses."""

    def reset_state(self) -> None:
        """Resets the state."""
        self._state = self._default_state

    def update_state(self, **kwargs: Any) -> None:
        """Pre-configures keyword arguments for featurise()."""
        self._state = {**self.state, **kwargs}



import pytest

import datasets
from molflux.datasets.load import load_dataset


def test_implicit_split_returns_single_dataset():
    """That using the default split value returns a single dataset."""
    dataset = load_dataset("esol")
    assert isinstance(dataset, datasets.Dataset)


def test_explicit_split_returns_dataset():
    """That using an explicit split value returns a single dataset."""
    dataset = load_dataset("esol", split="train")
    assert isinstance(dataset, datasets.Dataset)


def test_none_split_returns_dataset_dict():
    """That requesting a None split returns a dataset dict of all splits."""
    dataset = load_dataset("esol", split=None)
    assert isinstance(dataset, datasets.DatasetDict)


def test_bad_dataset_name_raises_an_error():
    """That a name not in the catalogue or online raises an error"""
    with pytest.raises(FileNotFoundError):
        load_dataset("molflux_certainly_not_a_dataset")



import pytest

from molflux.metrics import Metric, list_metrics, load_metric
from molflux.metrics.classification.matthews_corrcoef import MatthewsCorrcoef


@pytest.fixture()
def fixture_metric():
    return load_metric("matthews_corrcoef")


def test_metric_in_catalogue():
    """That the metric is registered in the catalogue."""
    catalogue = list_metrics()
    all_metric_names = [name for names in catalogue.values() for name in names]
    assert "matthews_corrcoef" in all_metric_names


def test_metric_is_mapped_to_correct_class(fixture_metric):
    """That the catalogue name is mapped to the appropriate class."""
    metric = fixture_metric
    assert isinstance(metric, MatthewsCorrcoef)


def test_implements_protocol(fixture_metric):
    """That the metric implements the public Metric protocol."""
    metric = fixture_metric
    assert isinstance(metric, Metric)


def test_integer_labels(fixture_metric):
    """That can use integer labels."""
    metric = fixture_metric
    predictions = [1, 1, 1, 0]
    references = [1, 0, 1, 1]
    result = metric.compute(predictions=predictions, references=references)
    assert result["matthews_corrcoef"] == -1 / 3


def test_perfect_score(fixture_metric):
    """That get unitary score for perfect predictions."""
    metric = fixture_metric
    predictions = [0, 1, 1, 0, 0, 1]
    references = [0, 1, 1, 0, 0, 1]
    result = metric.compute(predictions=predictions, references=references)
    assert result["matthews_corrcoef"] == 1.0


def test_random_score(fixture_metric):
    """That get null score for average random prediction."""
    metric = fixture_metric
    predictions = [0, 1, 1, 1, 1, 0]
    references = [0, 1, 1, 0, 0, 1]
    result = metric.compute(predictions=predictions, references=references)
    assert result["matthews_corrcoef"] == 0


def test_inverse_score(fixture_metric):
    """That get negative unitary score for inverse predictions."""
    metric = fixture_metric
    predictions = [1, 0, 0, 1, 1, 0]
    references = [0, 1, 1, 0, 0, 1]
    result = metric.compute(predictions=predictions, references=references)
    assert result["matthews_corrcoef"] == -1



import pandas as pd
import pytest

import datasets
from molflux.modelzoo import load_from_store, save_to_store
from molflux.modelzoo.catalogue import list_models
from molflux.modelzoo.load import load_model
from molflux.modelzoo.models.sklearn.logistic_regressor import LogisticRegressor
from molflux.modelzoo.protocols import Estimator, Model, supports_classification

model_name = "logistic_regressor"

_X_FEATURES = ["X_col_1", "X_col_2"]
_Y_FEATURES = ["y_col"]


train_df = pd.DataFrame(
    [
        [1, 2, 10],
        [1, 3, 11],
        [2, 4, 13],
    ],
    columns=_X_FEATURES + _Y_FEATURES,
)
predict_df = pd.DataFrame(
    [
        [1, 2],
        [1, 3],
        [2, 5],
    ],
    columns=_X_FEATURES,
)
empty_predict_df = pd.DataFrame([], columns=_X_FEATURES)


@pytest.fixture(scope="function")
def fixture_model() -> Model:
    return load_model(model_name, x_features=_X_FEATURES, y_features=_Y_FEATURES)


def test_in_catalogue():
    """That the model is registered in the catalogue."""
    catalogue = list_models()
    all_names = [name for names in catalogue.values() for name in names]
    assert model_name in all_names


def test_default_model_tag_matches_entrypoint_name(fixture_model):
    """That the default model tag matches the catalogue entrypoint name.

    This is not strictly required, but ensures a more intuitive user experience.
    """
    model = fixture_model
    assert model.tag == model_name


def test_is_mapped_to_correct_class(fixture_model):
    """That the model name is mapped to the appropriate class."""
    model = fixture_model
    assert isinstance(model, LogisticRegressor)


def test_implements_protocol(fixture_model):
    """That the model implements the public Estimator protocol."""
    model = fixture_model
    assert isinstance(model, Estimator)


def test_supports_classification(fixture_model):
    """That the model counts as being a classifier."""
    assert supports_classification(fixture_model)


@pytest.mark.parametrize(
    "train_data, predict_data",
    [
        (train_df, predict_df),
        (
            datasets.Dataset.from_pandas(train_df),
            datasets.Dataset.from_pandas(predict_df),
        ),
        (
            datasets.Dataset.from_pandas(train_df),
            datasets.Dataset.from_pandas(empty_predict_df),
        ),
    ],
)
def test_train_predict_model(fixture_model, train_data, predict_data):
    """Test that a model can run the train and predict functions"""

    model = fixture_model

    model.train(train_data)
    predictions = model.predict(predict_data)
    assert predictions
    assert len(predictions) == len(model.y_features)
    for task_predictions in predictions.values():
        assert len(task_predictions) == len(predict_data)


def test_predict_single_instance(fixture_model):
    """Test that predicting on a single data point behaves as expected"""
    train_data = datasets.Dataset.from_pandas(train_df)
    predict_data = datasets.Dataset.from_pandas(predict_df.iloc[:1])

    model = fixture_model
    model.train(train_data)

    predictions = model.predict(predict_data)
    for array in predictions.values():
        assert isinstance(array, list)


def test_saving_loading(tmp_path, fixture_model):
    """Test that a model can successfully be saved and loaded"""

    fixture_model.train(train_df)

    save_to_store(tmp_path, fixture_model)
    loaded_model = load_from_store(tmp_path)

    assert loaded_model.predict(predict_df)



import numpy as np
import scipy.stats as st

from molflux.metrics.typing import ArrayLike


def _estimate_standard_deviation(
    lower_bound: ArrayLike,
    upper_bound: ArrayLike,
    confidence: float,
) -> ArrayLike:
    # solve for standard deviation sigma based on difference between cdf at upper and lower bound
    # simpler estimate ignoring mu, and assuming symmetric gaussian distns for predictions
    def diff(x: ArrayLike) -> ArrayLike:
        return x[1] - x[0]

    prediction_intervals = list(zip(lower_bound, upper_bound, strict=False))
    sigma = [
        diff(interval) / diff(st.norm.interval(confidence))
        for interval in prediction_intervals
    ]
    return np.array(sigma)



from collections.abc import Iterable
from typing import Any, Literal

import pydantic
import yaml
from pydantic import BaseModel, Field

from molflux.metrics.typing import PathLike


class Spec(BaseModel):
    name: str
    config: dict[str, Any] = Field(default_factory=dict)
    presets: dict[str, Any] = Field(default_factory=dict)


class YamlConfig(BaseModel):
    version: str
    kind: Literal["metrics"]
    specs: list[dict[str, Any]]


def dict_parser(dictionary: dict[str, Any]) -> Spec:
    """Parses a dictionary into a spec."""
    try:
        return Spec(**dictionary)
    except pydantic.ValidationError as e:
        raise SyntaxError("Dictionary not conform with expected schema") from e


def yaml_parser(path: PathLike) -> Iterable[Spec]:
    """Parse a yaml file into specs."""

    with open(path) as f:
        documents = yaml.safe_load_all(f)

        # Support multiple documents in same file
        for document in documents:
            try:
                config = YamlConfig(**document)

                if config.version == "v1":
                    specs = [dict_parser(spec) for spec in config.specs]
                    return specs

                else:
                    raise NotImplementedError(
                        f"Yaml config schema not available: {config.version}",
                    )

            except pydantic.ValidationError:
                pass

    raise FileNotFoundError(f"Could not find valid document in file: {path}")



"""
Abstract Base Classes for classes implementing the Model protocol.
"""

import inspect
import logging
import os
import pickle
import warnings
from abc import ABC, abstractmethod
from dataclasses import field
from functools import cached_property
from typing import Any, Generic, TypeVar

from pydantic.v1 import dataclasses

from datasets import Dataset, concatenate_datasets
from molflux.modelzoo.errors import NotTrainedError
from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.interchange import hf_dataset_from_dataframe
from molflux.modelzoo.naming import camelcase_to_snakecase
from molflux.modelzoo.typing import (
    Classes,
    DataFrameLike,
    Features,
    PredictionResult,
)
from molflux.modelzoo.utils import (
    pick_features,
    raise_on_unrecognised_parameters,
    validate_features,
    validate_prediction_result_num_tasks,
)
from molflux.version import version as __version__

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class ModelConfig:
    x_features: Features = field(default_factory=list)
    y_features: Features = field(default_factory=list)
    train_features: Features | dict[str, Features] | None = None

    def resolve_train_features(self, name: str | None = None) -> Features:
        """Returns the features expected in training dataset `name`."""
        if self.train_features is None:
            return self.x_features + self.y_features
        elif isinstance(self.train_features, dict):
            if name is None:
                raise ValueError(
                    "This model's features differ by dataset. Specify its name.",
                )

            return self.train_features.get(name, self.x_features + self.y_features)
        else:
            return self.train_features


_ModelConfigT = TypeVar("_ModelConfigT", bound=ModelConfig)


class ModelBase(Generic[_ModelConfigT], ABC):
    _train_multi_data_enabled: bool = False

    def __init__(self, *, tag: str | None = None, **config_kwargs: Any) -> None:
        """Initialises a new model."""

        # build default config
        self.model_config = self._config_builder(**config_kwargs)

        # build default info (with custom values if specified)
        info = self._info()
        info.name = camelcase_to_snakecase(type(self).__name__)
        info.tag = tag or info.name
        info.config = self.config
        info.version = __version__
        self.info = info

        # The backend model object(s)
        self.model: Any = None

    @property
    def config(self) -> dict[str, Any]:
        return self.model_config.__dict__

    @cached_property
    def config_signature(self) -> inspect.Signature:
        """The model config signature.

        This is defined as a property to work on
        unpickled objects from previous modelzoo versions.
        """
        return inspect.signature(self.model_config.__class__.__init__)

    @cached_property
    def predict_signature(self) -> inspect.Signature:
        """The predict signature.

        This is defined as a property to be able to run predictions
        on unpickled objects from previous modelzoo versions."""
        return inspect.signature(self._predict)

    @cached_property
    def train_signature(self) -> inspect.Signature:
        """The train signature.

        This is defined as a property to be able to train
        unpickled objects from previous modelzoo versions."""
        return inspect.signature(self._train)

    @property
    def x_features(self) -> Features:
        return self.model_config.x_features

    @property
    def y_features(self) -> Features:
        return self.model_config.y_features

    @property
    def train_features(self) -> Features | dict[str, Features] | None:
        return self.model_config.train_features

    def __str__(self) -> str:
        return (
            f"Model(\n"
            f'\tname: "{self.name}",\n'
            f'\ttag: "{self.tag}",\n'
            f'\tdescription: """{self.info.model_description}""",\n'
            f"\tconfig signature: __init__{self.config_signature},\n"
            f'\tconfig: """{self.info.config_description}""",\n'
            f"\ttrain signature: self.train{self.train_signature},\n"
            f"\tpredict signature: self.predict{self.predict_signature},\n"
            ")"
        )

    @property
    @abstractmethod
    def _config_builder(self) -> type[_ModelConfigT]:
        """The callable that initialises the model config object.

        To be implemented by subclasses.
        """

    @abstractmethod
    def _info(self) -> ModelInfo:
        """Initialises the ModelInfo object.

        To be implemented by subclasses.
        """

    @property
    def metadata(self) -> dict[str, Any]:
        return self.info.to_dict()

    @property
    def name(self) -> str:
        return self.info.name

    @property
    def tag(self) -> str:
        return self.info.tag

    def train(
        self,
        train_data: DataFrameLike | dict[str | None, DataFrameLike],
        validation_data: DataFrameLike | dict[str | None, DataFrameLike] | None = None,
        **kwargs: Any,
    ) -> Any:
        """Trains the model."""

        all_data = {"train_data": train_data}

        if self._train_multi_data_enabled:
            # safe support for optional validation_data used e.g. by Lightning models
            if validation_data is not None:
                try:
                    raise_on_unrecognised_parameters(
                        self._train_multi_data,
                        validation_data=validation_data,
                    )
                    all_data["validation_data"] = validation_data
                except ValueError:
                    warnings.warn(
                        "Validation data not supported for this model type. It will not propagate to _train_multi_data.",
                        UserWarning,
                        stacklevel=2,
                    )

            all_multi_datasets: dict[str, dict[str | None, Dataset]] = {}
            for split, data in all_data.items():
                # if passed a single dataset, convert to dict
                if not isinstance(data, dict):
                    data = {None: data}

                # make sure data is a datasets.Dataset
                datasets = {k: hf_dataset_from_dataframe(v) for k, v in data.items()}

                # make sure dataset has required features
                # NOTE ensures that train_features are present in the validation dataset, too
                for name, dataset in datasets.items():
                    validate_features(
                        dataset,
                        self.model_config.resolve_train_features(name),
                    )

                all_multi_datasets[split] = datasets

            # Safeguard against invalid kwargs
            raise_on_unrecognised_parameters(self._train_multi_data, **kwargs)

            return self._train_multi_data(**all_multi_datasets, **kwargs)

        else:
            # safe support for optional validation_data used e.g. by Lightning models
            if validation_data is not None:
                try:
                    raise_on_unrecognised_parameters(
                        self._train,
                        validation_data=validation_data,
                    )
                    all_data["validation_data"] = validation_data
                except ValueError:
                    warnings.warn(
                        "Validation data not supported for this model type. It will not propagate to _train.",
                        UserWarning,
                        stacklevel=2,
                    )

            all_datasets: dict[str, Dataset] = {}
            for split, data in all_data.items():
                if isinstance(data, dict):
                    dataset = concatenate_datasets(
                        [
                            hf_dataset_from_dataframe(dataset)
                            for dataset in data.values()
                        ],
                    )
                else:
                    dataset = hf_dataset_from_dataframe(data)

                # NOTE ensures that train_features are present in the validation dataset, too
                validate_features(dataset, self.model_config.resolve_train_features())

                all_datasets[split] = dataset

            raise_on_unrecognised_parameters(self._train, **kwargs)

            return self._train(**all_datasets, **kwargs)

    @abstractmethod
    def _train(self, train_data: Dataset, **kwargs: Any) -> Any:
        """The training callable.

        To be implemented by subclasses.
        """

    def _train_multi_data(
        self,
        train_data: dict[str | None, Dataset],
        **kwargs: Any,
    ) -> Any:
        """The training callable for multiple datasets.

        To be implemented by subclasses if they support multiple datasets.
        """
        del train_data, kwargs
        raise NotImplementedError(
            "This class does not support training on multiple datasets.",
        )

    def predict(self, data: DataFrameLike, **kwargs: Any) -> PredictionResult:
        """Performs a model prediction."""

        # make sure data is a datasets.Dataset
        data = hf_dataset_from_dataframe(data)

        # make sure dataset has required features
        validate_features(data, self.x_features)

        # pick out relevant features
        data = pick_features(data, self.x_features)

        # Safeguard against invalid kwargs
        raise_on_unrecognised_parameters(self._predict, **kwargs)

        prediction_result = self._predict(data=data, **kwargs)

        validate_prediction_result_num_tasks(prediction_result, self.y_features)

        return prediction_result

    @abstractmethod
    def _predict(self, data: Dataset, **kwargs: Any) -> PredictionResult:
        """The prediction callable.

        To be implemented by subclasses.
        """

    @property
    def _predict_display_names(self) -> list[str]:
        """The display names for prediction outputs.

        These are not part of the API, we just provide this utility function
        to help making display names consistent across all of our models.
        """
        return [f"{self.tag}::{task}" for task in self.y_features]

    def as_dir(self, directory: str) -> None:
        """Serialises a pre-trained model in a directory."""

        if self.model is None:
            raise NotTrainedError

        pickle_fn = os.path.join(directory, "model.pkl")
        with open(pickle_fn, "wb") as f:
            pickle.dump(self.model, f)

    def from_dir(self, directory: str) -> None:
        """Deserialises the backend model object stored in a given directory.

        This should undo the process defined in `self.as_dir()`.
        """

        # The expected model binary
        filename = os.path.join(directory, "model.pkl")
        with open(filename, "rb") as f:
            self.model = pickle.load(f)  # noqa: S301
            return


class ClassificationMixin(ABC):
    """Mixin for all classifiers in molflux-modelzoo."""

    tag: str
    x_features: Features
    y_features: Features

    @property
    @abstractmethod
    def classes(self) -> Classes:
        """Dictionary mapping each task name to a list of possible classes"""
        ...

    @cached_property
    def predict_proba_signature(self) -> inspect.Signature:
        """The predict_proba signature."""
        return inspect.signature(self._predict_proba)

    def predict_proba(self, data: DataFrameLike, **kwargs: Any) -> PredictionResult:
        """Returns probability estimates for the data."""

        # make sure data is a datasets.Dataset
        data = hf_dataset_from_dataframe(data)

        # make sure dataset has required features
        validate_features(data, self.x_features)

        # pick out relevant features
        data = pick_features(data, self.x_features)

        # Safeguard against invalid kwargs
        raise_on_unrecognised_parameters(self._predict_proba, **kwargs)

        prediction_result = self._predict_proba(data=data, **kwargs)

        validate_prediction_result_num_tasks(prediction_result, self.y_features)

        return prediction_result

    @abstractmethod
    def _predict_proba(self, data: Dataset, **kwargs: Any) -> PredictionResult:
        """The predict_proba callable.

        To be implemented by subclasses.
        """

    @property
    def _predict_proba_display_names(self) -> list[str]:
        """The display names for prediction outputs.

        These are not part of the API, we just provide this utility function
        to help making display names consistent across all of our models.
        """
        return [f"{self.tag}::{task}::probabilities" for task in self.y_features]

    def __str__(self) -> str:
        return super().__str__() + (
            f"\nClassifier(\n"
            f'\tclasses: "{self.classes}",\n'
            f"\tpredict_proba signature: self.predict_proba{self.predict_proba_signature},\n"
            ")"
        )


class UncertaintyCalibrationMixin(ABC):
    """Mixin for all models supporting uncertainty calibration molflux-modelzoo"""

    x_features: Features
    y_features: Features

    @cached_property
    def calibrate_uncertainty_signature(self) -> inspect.Signature:
        """The calibrate_uncertainty signature."""
        return inspect.signature(self._calibrate_uncertainty)

    def calibrate_uncertainty(self, data: DataFrameLike, **kwargs: Any) -> Any:
        """Calibrates the uncertainty of predictions on a validation dataset"""

        # make sure data is a datasets.Dataset
        data = hf_dataset_from_dataframe(data)

        # make sure dataset has required features
        # y features are not validated as not all models might define them
        validate_features(data, self.x_features)

        # Safeguard against invalid kwargs
        raise_on_unrecognised_parameters(self._calibrate_uncertainty, **kwargs)

        # idea: could rearrange here the predictions and names depending on the format
        return self._calibrate_uncertainty(data=data, **kwargs)

    @abstractmethod
    def _calibrate_uncertainty(self, data: Dataset, **kwargs: Any) -> Any:
        """The uncertainty calibration callable.

        To be implemented by subclasses.
        """

    def __str__(self) -> str:
        return super().__str__() + (
            f"\nEstimator(\n"
            f"\tcalibrate_uncertainty signature: self.calibrate_uncertainty_signature{self.calibrate_uncertainty_signature},\n"
            ")"
        )


class PredictionIntervalMixin(ABC):
    """Mixin for all models supporting prediction intervals."""

    tag: str
    x_features: Features
    y_features: Features

    @cached_property
    def predict_with_prediction_interval_signature(self) -> inspect.Signature:
        return inspect.signature(self._predict_with_prediction_interval)

    def predict_with_prediction_interval(
        self,
        data: DataFrameLike,
        confidence: float,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        """Return a model prediction with an associated prediction interval"""

        # make sure data is a datasets.Dataset
        data = hf_dataset_from_dataframe(data)

        # make sure dataset has required features
        validate_features(data, self.x_features)

        # pick out relevant features
        data = pick_features(data, self.x_features)

        # Safeguard against invalid kwargs
        raise_on_unrecognised_parameters(
            self._predict_with_prediction_interval,
            **kwargs,
        )

        (
            prediction_result,
            prediction_prediction_interval_result,
        ) = self._predict_with_prediction_interval(
            data,
            confidence=confidence,
            **kwargs,
        )

        validate_prediction_result_num_tasks(prediction_result, self.y_features)
        validate_prediction_result_num_tasks(
            prediction_prediction_interval_result,
            self.y_features,
        )

        return prediction_result, prediction_prediction_interval_result

    @abstractmethod
    def _predict_with_prediction_interval(
        self,
        data: Dataset,
        confidence: float,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        """The prediction interval prediction callable for prediction intervals.

        To be implemented by subclasses.
        """

    @property
    def _predict_with_prediction_interval_display_names(
        self,
    ) -> tuple[list[str], list[str]]:
        """The display names for prediction outputs.

        These are not part of the API, we just provide this utility function
        to help making display names consistent across all of our models.
        """
        return zip(  # type: ignore[return-value]
            *[
                (f"{self.tag}::{task}", f"{self.tag}::{task}::prediction_interval")
                for task in self.y_features
            ],
            strict=False,
        )

    def __str__(self) -> str:
        return super().__str__() + (
            f"\nEstimator(\n"
            f"\tpredict_with_prediction_interval signature: self.predict_with_prediction_interval{self.predict_with_prediction_interval},\n"
            ")"
        )


class StandardDeviationMixin(ABC):
    """Mixin for all models supporting predictions with standard deviations"""

    tag: str
    x_features: Features
    y_features: Features

    @cached_property
    def predict_with_std_signature(self) -> inspect.Signature:
        return inspect.signature(self._predict_with_std)

    def predict_with_std(
        self,
        data: DataFrameLike,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        """Return a model prediction with an associated standard deviation"""

        # make sure data is a datasets.Dataset
        data = hf_dataset_from_dataframe(data)

        # make sure dataset has required features
        validate_features(data, self.x_features)

        # pick out relevant features
        data = pick_features(data, self.x_features)

        # Safeguard against invalid kwargs
        raise_on_unrecognised_parameters(self._predict_with_std, **kwargs)

        prediction_result, prediction_std_result = self._predict_with_std(
            data,
            **kwargs,
        )

        validate_prediction_result_num_tasks(prediction_result, self.y_features)
        validate_prediction_result_num_tasks(prediction_std_result, self.y_features)

        return prediction_result, prediction_std_result

    @abstractmethod
    def _predict_with_std(
        self,
        data: Dataset,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        """The prediction callable for prediction standard deviations.

        To be implemented by subclasses.
        """

    @property
    def _predict_with_std_display_names(self) -> tuple[list[str], list[str]]:
        """The display names for prediction outputs.

        These are not part of the API, we just provide this utility function
        to help making display names consistent across all of our models.
        """
        return zip(  # type: ignore[return-value]
            *[
                (f"{self.tag}::{task}", f"{self.tag}::{task}::std")
                for task in self.y_features
            ],
            strict=False,
        )

    def __str__(self) -> str:
        return super().__str__() + (
            f"\nEstimator(\n"
            f"\tpredict_with_std signature: self.predict_with_std{self.predict_with_std_signature},\n"
            ")"
        )


class CovarianceMixin(ABC):
    """Mixin for all models supporting predictions with a covariance matrix of joint predictions"""

    tag: str
    x_features: Features
    y_features: Features

    @cached_property
    def predict_with_covariance_signature(self) -> inspect.Signature:
        return inspect.signature(self._predict_with_covariance)

    def predict_with_covariance(
        self,
        data: DataFrameLike,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        """Return a model prediction with an associated standard deviation"""

        # make sure data is a datasets.Dataset
        data = hf_dataset_from_dataframe(data)

        # make sure dataset has required features
        validate_features(data, self.x_features)

        # pick out relevant features
        data = pick_features(data, self.x_features)

        # Safeguard against invalid kwargs
        raise_on_unrecognised_parameters(self._predict_with_covariance, **kwargs)

        prediction_result, prediction_covariance_result = self._predict_with_covariance(
            data,
            **kwargs,
        )

        validate_prediction_result_num_tasks(prediction_result, self.y_features)
        validate_prediction_result_num_tasks(
            prediction_covariance_result,
            self.y_features,
        )

        return prediction_result, prediction_covariance_result

    @abstractmethod
    def _predict_with_covariance(
        self,
        data: Dataset,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        """The prediction callable for prediction covariances.

        To be implemented by subclasses.
        """

    @property
    def _predict_with_covariance_display_names(self) -> tuple[list[str], list[str]]:
        """The display names for prediction outputs.

        These are not part of the API, we just provide this utility function
        to help making display names consistent across all of our models.
        """
        return zip(  # type: ignore[return-value]
            *[
                (f"{self.tag}::{task}", f"{self.tag}::{task}::covariance")
                for task in self.y_features
            ],
            strict=False,
        )

    def __str__(self) -> str:
        return super().__str__() + (
            f"\nEstimator(\n"
            f"\tpredict_with_covariance signature: self.predict_with_covariance{self.predict_with_covariance_signature},\n"
            ")"
        )


class SamplingMixin(ABC):
    """Mixin for all models supporting sampling"""

    tag: str
    x_features: Features
    y_features: Features

    @cached_property
    def sample_signature(self) -> inspect.Signature:
        return inspect.signature(self._sample)

    def sample(
        self,
        data: DataFrameLike,
        n_samples: int,
        **kwargs: Any,
    ) -> PredictionResult:
        """Return n_samples prediction estimates with the model"""

        # make sure data is a datasets.Dataset
        data = hf_dataset_from_dataframe(data)

        # make sure dataset has required features
        validate_features(data, self.x_features)

        # pick out relevant features
        data = pick_features(data, self.x_features)

        # Safeguard against invalid kwargs
        raise_on_unrecognised_parameters(self._sample, n_samples=n_samples, **kwargs)

        sampling_result = self._sample(data, n_samples=n_samples, **kwargs)

        validate_prediction_result_num_tasks(sampling_result, self.y_features)

        return sampling_result

    @abstractmethod
    def _sample(
        self,
        data: Dataset,
        n_samples: int,
        **kwargs: Any,
    ) -> PredictionResult:
        """The samples callable, generating n_samples predictions for the input.

        To be implemented by subclasses.
        """

    @property
    def _sample_display_names(self) -> list[str]:
        """The display names for prediction outputs.

        These are not part of the API, we just provide this utility function
        to help making display names consistent across all of our models.
        """
        return [f"{self.tag}::{task}::samples" for task in self.y_features]

    def __str__(self) -> str:
        return super().__str__() + (
            f"\nEstimator(\n"
            f"\tsample signature: self.sample{self.sample_signature},\n"
            ")"
        )



from molflux.metrics import load_metric
from molflux.metrics.protocols import supports_prediction_intervals


def test_supports_prediction_intervals_positive_example():
    """this example should support uncertainty"""
    metric = load_metric(name="gaussian_nll")
    assert supports_prediction_intervals(metric)


def test_supports_prediction_intervals_negative_example():
    """this example should not support uncertainty"""
    metric = load_metric(name="spearman")
    assert not supports_prediction_intervals(metric)



import warnings
from copy import copy
from typing import Any, Final, Literal

import numpy as np
from pydantic.v1 import dataclasses
from scipy.stats import norm

import datasets
from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.load import load_from_dict
from molflux.modelzoo.model import (
    ModelBase,
    ModelConfig,
    PredictionIntervalMixin,
    SamplingMixin,
    StandardDeviationMixin,
    UncertaintyCalibrationMixin,
)
from molflux.modelzoo.models.sklearn import SKLearnModelBase
from molflux.modelzoo.protocols import Estimator
from molflux.modelzoo.typing import PredictionResult
from molflux.modelzoo.utils import (
    format_wrapped_model_tag,
    get_concatenated_array,
    pick_features,
    validate_features,
)

try:
    from mapie.conformity_scores import (
        AbsoluteConformityScore,
        GammaConformityScore,
        ResidualNormalisedScore,
    )
    from mapie.regression import MapieRegressor as MapieMapieRegressor
    from sklearn.base import RegressorMixin
    from sklearn.model_selection import BaseCrossValidator

except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("mapie", e) from None


_DESCRIPTION = """
Prediction interval with out-of-fold conformity scores.

This class implements the jackknife+ strategy and its variations
for estimating prediction intervals on single-output data. The
idea is to evaluate out-of-fold conformity scores (signed residuals,
absolute residuals, residuals normalized by the predicted mean...)
on hold-out validation sets and to deduce valid prediction intervals
with strong theoretical guarantees.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
estimator : Union[Estimator,RegressorMixin,None,Dict]
    Any modelzoo estimator or sklearn regressor backed by a scikit-learn API
    (i.e. with fit and predict methods), by default ``None``.
    If ``None``, estimator defaults to a linear regressor instance.
    If a dictionary, it is assumed to me a modelzoo model config for an scikit-learn backed regressor.

method: str, optional
    Method to choose for prediction interval estimates.
    Choose among:

    - "naive", based on training set conformity scores,
    - "base", based on validation sets conformity scores,
    - "plus", based on validation conformity scores and
      testing predictions,
    - "minmax", based on validation conformity scores and
      testing predictions (min/max among cross-validation clones).

    By default "plus".

cv: Optional[Union[int, str, BaseCrossValidator]]
    The cross-validation strategy for computing conformity scores.
    It directly drives the distinction between jackknife and cv variants.
    Choose among:

    - ``None``, to use the default 5-fold cross-validation
    - integer, to specify the number of folds.
      If equal to -1, equivalent to
      ``sklearn.model_selection.LeaveOneOut()``.
    - CV splitter: any ``sklearn.model_selection.BaseCrossValidator``
      Main variants are:
      - ``sklearn.model_selection.LeaveOneOut`` (jackknife),
      - ``sklearn.model_selection.KFold`` (cross-validation),
      - ``subsample.Subsample`` object (bootstrap).
    - ``"prefit"``, assumes that ``estimator`` has been fitted already,
      and the ``method`` parameter is ignored.
      All data provided in the ``fit`` method is then used
      for computing conformity scores only.
      At prediction time, quantiles of these conformity scores are used
      to provide a prediction interval with fixed width.
      The user has to take care manually that data for model fitting and
      conformity scores estimate are disjoint.

    By default ``None``.

n_jobs: Optional[int]
    Number of jobs for parallel processing using joblib
    via the "locky" backend.
    If ``-1`` all CPUs are used.
    If ``1`` is given, no parallel computing code is used at all,
    which is useful for debugging.
    For n_jobs below ``-1``, ``(n_cpus + 1 - n_jobs)`` are used.
    None is a marker for `unset` that will be interpreted as ``n_jobs=1``
    (sequential execution).

    By default ``None``.

agg_function : str
    Determines how to aggregate predictions from perturbed models, both at
    training and prediction time.

    If ``None``, it is ignored except if cv class is ``Subsample``,
    in which case an error is raised.
    If "mean" or "median", returns the mean or median of the predictions
    computed from the out-of-folds models.
    Note: if you plan to set the ``ensemble`` argument to ``True`` in the
    ``predict`` method, you have to specify an aggregation function.
    Otherwise an error would be raised.

    The Jackknife+ interval can be interpreted as an interval around the
    median prediction, and is guaranteed to lie inside the interval,
    unlike the single estimator predictions.

    When the cross-validation strategy is Subsample (i.e. for the
    Jackknife+-after-Bootstrap method), this function is also used to
    aggregate the training set in-sample predictions.

    If cv is ``"prefit"``, ``agg_function`` is ignored.

    By default "mean".

verbose : int, optional
    The verbosity level, used with joblib for multiprocessing.
    The frequency of the messages increases with the verbosity level.
    If it more than ``10``, all iterations are reported.
    Above ``50``, the output is sent to stdout.

    By default ``0``.

conformity_score: Optional[Union[str, ConformityScore]]
    ConformityScore instance.
    It defines the link between the observed values, the predicted ones
    and the conformity scores. For instance, the default ``None`` value
    correspondonds to a conformity score which assumes
    y_obs = y_pred + conformity_score.

    - default is ``AbsoluteConformityScore`` conformity
      score
    - ConformityScore: any ``ConformityScore`` class
    - Currently only enable AbsoluteConformityScore, GammaConformityScore,
      or ResidualNormalisedScore
    - with ResidualNormalisedScore, normalisation is by an sklearn model -
      at the moment, don't allow to configure this - use default linear regressor
"""

Method = Literal["naive", "base", "plus", "minmax"]
AggFunction = Literal["mean", "median"]
Score = Literal[
    "AbsoluteConformityScore",
    "GammaConformityScore",
    "ResidualNormalisedScore",
]

_UNLINKED_FLAG: Final = "UNLINKED"


def _as_mapie_estimator(model: Any) -> Any:
    """Converts a generic model into an estimator suitable for mapie.MapieRegressor.

    This allows us to expand the existing mapie interface to accept other inputs,
    such as molflux-modelzoo estimators.

    Returns:
        An estimator that is compatible with the interface expected by mapie.
    """
    if model == _UNLINKED_FLAG:
        raise ValueError(
            "The input estimator has been unlinked. This is likely to have happened because you are reusing a pre-trained MAPIE model. Please train a new model from scratch instead.",
        )

    # Pass-through the native types accepted by mapie.MapieRegressor
    if model is None or isinstance(model, RegressorMixin):
        return model

    # a dictionary is assumed to be a modelzoo config
    if isinstance(model, dict):
        try:
            model = load_from_dict(model)
        except Exception as e:
            raise ValueError(
                f"Could not generate a modelzoo model from the estimator config {model}",
            ) from e

    # Allow to pass generic modelzoo architectures as well
    # TODO(avianello): for the time being we will only allow modezloo models
    #  explicitly backed by a sklearn model. In the future we could expand by
    #  creating an adapter class to turn any modelzoo estimator into the
    #  interface accepted by mapie
    if isinstance(model, Estimator):
        if isinstance(model, SKLearnModelBase):
            # warning: this is not a method provided by the Estimator Protocol
            # it needs to be called because the model.model attribute is None
            # before fitting a modelzoo model
            if model.model is None:
                return model._instantiate_model()
            return model.model
        raise NotImplementedError(
            f"Unsupported estimator architecture: {model.__class__!r}",
        )

    raise TypeError(f"Invalid estimator type for MapieRegressor: {type(model)!r}")


class Config:
    arbitrary_types_allowed = True


@dataclasses.dataclass(config=Config)
class MapieRegressorConfig(ModelConfig):
    estimator: Estimator | RegressorMixin | None | str | dict = None
    method: Method = "plus"
    cv: int | str | BaseCrossValidator | None = None
    n_jobs: int | None = None
    agg_function: AggFunction | None = "mean"
    verbose: int = 0
    conformity_score: Score | None = "AbsoluteConformityScore"

    def __post_init_post_parse__(self) -> None:
        if self.y_features and len(self.y_features) != 1:
            raise NotImplementedError(
                f"This model architecture only supports single task regression for now: got {self.y_features}",
            )

        if self.estimator == _UNLINKED_FLAG:
            warnings.warn(
                "Model loaded with an unlinked input estimator. This is expected if loading a pre-trained MAPIE model. Note that the resulting model can only be used for prediction and cannot be retrained.",
                UserWarning,
                stacklevel=1,
            )


class MapieRegressor(
    UncertaintyCalibrationMixin,
    PredictionIntervalMixin,
    StandardDeviationMixin,
    SamplingMixin,
    ModelBase[MapieRegressorConfig],
):
    def __init__(self, tag: str | None = None, **config_kwargs: Any) -> None:
        super().__init__(tag=tag, **config_kwargs)
        estimator = self.model_config.estimator
        if isinstance(estimator, Estimator):
            base_estimator_tag = estimator.tag
        elif isinstance(estimator, RegressorMixin):
            base_estimator_tag = estimator.__name__
        else:
            base_estimator_tag = "linear_regressor"
        if tag is None:
            self.info.tag = format_wrapped_model_tag(
                self.info.tag,
                [base_estimator_tag],
            )

    @property
    def _config_builder(self) -> type[MapieRegressorConfig]:
        return MapieRegressorConfig

    @property
    def config(self) -> dict[str, Any]:
        # handle unserializable config fields
        config = copy(self.model_config.__dict__)
        config["estimator"] = _UNLINKED_FLAG
        return config

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> MapieMapieRegressor:
        config = self.model_config

        # the conformity score is created on the fly
        if config.conformity_score == "GammaConformityScore":
            conformity_score = GammaConformityScore()
        elif config.conformity_score == "ResidualNormalisedScore":
            conformity_score = ResidualNormalisedScore()
        else:
            conformity_score = AbsoluteConformityScore()

        # avoids automatic checks that sometimes produce errors depending on
        # the dataset and machine due to machine decimal errors
        # when using the conformal methods provided by mapie, there are
        # mathematical guarantees that such checks are redundant
        conformity_score.consistency_check = False

        mapie_regressor = MapieMapieRegressor(
            estimator=_as_mapie_estimator(config.estimator),
            method=config.method,
            cv=config.cv,
            n_jobs=config.n_jobs,
            agg_function=config.agg_function,
            verbose=config.verbose,
            conformity_score=conformity_score,
        )
        return mapie_regressor

    def _train(
        self,
        train_data: datasets.Dataset,
        **kwargs: Any,
    ) -> Any:
        # validate y features as well
        validate_features(train_data, self.y_features)

        x_data = pick_features(train_data, self.x_features)
        X = get_concatenated_array(x_data, self.x_features)

        y_data = pick_features(train_data, self.y_features)
        y = get_concatenated_array(y_data, self.y_features)

        # instantiate model
        self.model = self._instantiate_model()

        # train
        self.model.fit(X, y, **kwargs)

    def _predict(
        self,
        data: datasets.Dataset,
        use_ensemble_predictions: bool = True,
        **kwargs: Any,
    ) -> PredictionResult:
        # the conformal predictions require an alpha, which is 1-confidence, between 1/n and (n-1)/n
        # here, we'll 1/2 which respects that for any n>=2
        default_confidence = 0.5

        prediction_result, _ = self._predict_with_prediction_interval(
            data,
            confidence=default_confidence,
            use_ensemble_predictions=use_ensemble_predictions,
            **kwargs,
        )
        return prediction_result

    def _calibrate_uncertainty(self, data: datasets.Dataset, **kwargs: Any) -> Any:
        return self._train(train_data=data, **kwargs)

    def _predict_with_std(
        self,
        data: datasets.Dataset,
        use_ensemble_predictions: bool = True,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        (
            _,
            prediction_interval_display_names,
        ) = self._predict_with_prediction_interval_display_names
        (
            _,
            prediction_std_display_names,
        ) = self._predict_with_std_display_names
        # Predict a 1-sigma prediction interval. Use this to provide a gaussian approximation
        pred_output, pred_interval_output = self._predict_with_prediction_interval(
            data=data,
            confidence=norm.cdf(1) - norm.cdf(-1),  # approximately 0.6827
            use_ensemble_predictions=use_ensemble_predictions,
            **kwargs,
        )
        pred_std_output: PredictionResult = {}
        for interval_display_name, std_display_name in zip(
            prediction_interval_display_names,
            prediction_std_display_names,
            strict=False,
        ):
            pred_std_output[std_display_name] = [
                (ub - lb) / 2.0
                for lb, ub in pred_interval_output[interval_display_name]
            ]

        return pred_output, pred_std_output

    def _sample(
        self,
        data: datasets.Dataset,
        n_samples: int,
        use_ensemble_predictions: bool = True,
        **kwargs: Any,
    ) -> PredictionResult:
        display_names = self._sample_display_names

        if not len(data):
            return {display_name: [] for display_name in display_names}

        prediction_mean_results, prediction_std_results = self._predict_with_std(
            data=data,
            use_ensemble_predictions=use_ensemble_predictions,
            **kwargs,
        )

        prediction_results: PredictionResult = {}
        for display_name, means, stds in zip(
            display_names,
            prediction_mean_results.values(),
            prediction_std_results.values(),
            strict=False,
        ):
            samples = np.random.normal(means, stds, (n_samples, len(means))).T
            prediction_results[display_name] = samples.tolist()

        return prediction_results

    def _predict_with_prediction_interval(
        self,
        data: datasets.Dataset,
        confidence: float,
        use_ensemble_predictions: bool = True,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        (
            prediction_display_names,
            prediction_interval_display_names,
        ) = self._predict_with_prediction_interval_display_names

        if not len(data):
            return {display_name: [] for display_name in prediction_display_names}, {
                display_name: [] for display_name in prediction_interval_display_names
            }

        X = get_concatenated_array(data, self.x_features)

        # Evaluate prediction on testing set
        y_predict, y_pis = self.model.predict(
            X,
            ensemble=use_ensemble_predictions,
            alpha=1 - confidence,
        )
        lower_bound, upper_bound = y_pis[:, 0, 0], y_pis[:, 1, 0]

        # Check that predictions match expected features
        y_features = self.y_features
        if y_predict.ndim == 1:
            if len(y_features) != 1:
                raise ValueError(
                    f"Predictions do not have expected shape: {np.shape(y_predict)!r}",
                )
        elif y_predict.shape[1] != len(y_features):
            raise ValueError(
                f"Predictions do not have expected shape: {np.shape(y_predict)!r}",
            )

        # this model architecture only supports single task models for now
        return {
            display_name: y_predict.tolist()
            for display_name in prediction_display_names
        }, {
            display_name: list(zip(lower_bound, upper_bound, strict=False))
            for display_name in prediction_interval_display_names
        }



import pytest

from molflux.features import Representation, list_representations, load_representation
from molflux.features.representations.openeye.shape.hermite import (
    Hermite,
)

representation_name = "hermite"


@pytest.fixture(scope="function")
def fixture_representation() -> Representation:
    return load_representation(representation_name)


def test_representation_in_catalogue():
    """That the representation is registered in the catalogue."""
    catalogue = list_representations()
    all_representation_names = [name for names in catalogue.values() for name in names]
    assert representation_name in all_representation_names


def test_representation_is_mapped_to_correct_class(fixture_representation):
    """That the catalogue name is mapped to the appropriate class."""
    representation = fixture_representation
    assert isinstance(representation, Hermite)


def test_implements_protocol(fixture_representation):
    """That the representation implements the public Representation protocol."""
    representation = fixture_representation
    assert isinstance(representation, Representation)


def test_default_compute(fixture_representation):
    """That default scoring gives expected results."""
    representation = fixture_representation
    samples = [
        "CCCC1=NN(C2=C1NC(=NC2=O)C3=C(C=CC(=C3)S(=O)(=O)N4CCN(CC4)C)OCC)C",
    ]
    result = representation.featurise(samples, n_poly_max=2)
    expected_result = [
        [
            89.11933047322135,
            0.0,
            0.003222152914168836,
            0.0,
            0.0,
            0.0032220784927158943,
            0.0,
            0.0,
            0.0,
            0.003221965510728286,
        ],
    ]
    assert representation_name in result
    assert result[representation_name][0] == pytest.approx(
        expected_result[0],
        abs=0.0005,
    )


def test_batch_compute(fixture_representation):
    """That batch scoring gives expected results."""
    representation = fixture_representation
    samples = [
        "CCCC1=NN(C2=C1NC(=NC2=O)C3=C(C=CC(=C3)S(=O)(=O)N4CCN(CC4)C)OCC)C",
        "CC",
    ]
    result = representation.featurise(samples, n_poly_max=2)
    expected_results = [
        [
            89.11933047322135,
            0.0,
            0.003222152914168836,
            0.0,
            0.0,
            0.0032220784927158943,
            0.0,
            0.0,
            0.0,
            0.003221965510728286,
        ],
        [
            5.3907372015426205,
            0.0,
            -0.0015415925639253993,
            0.0,
            0.0,
            -0.0015415925424404435,
            0.0,
            0.0,
            0.0,
            -0.0015415925330217136,
        ],
    ]
    assert representation_name in result
    for j, exp_res in enumerate(expected_results):
        assert result[representation_name][j] == pytest.approx(exp_res, abs=0.0005)



from typing import Any, Literal, Optional, Union

from numpy.random import RandomState
from pydantic.v1 import dataclasses

import datasets
from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.sklearn import (
    SKLearnClassificationMixin,
    SKLearnModelBase,
)
from molflux.modelzoo.typing import PredictionResult

try:
    from sklearn.svm import SVC
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from None


_DESCRIPTION = """
This is an sklearn support vector classifier model.

The objective of a support vector machine algorithm is to find a hyperplane in an
n-dimensional space that best separates classes of the data. The data points on
either side of the hyperplane that are closest to the hyperplane are called Support
Vectors. These influence the position and orientation of the hyperplane and thus help
build the SVM.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
C : float, default=1.0
    Regularization parameter. The strength of the regularization is
    inversely proportional to C. Must be strictly positive. The penalty
    is a squared l2 penalty.
kernel : {'linear', 'poly', 'rbf', 'sigmoid', 'precomputed'} or callable,  \
    default='rbf'
    Specifies the kernel type to be used in the algorithm.
    If none is given, 'rbf' will be used. If a callable is given it is
    used to pre-compute the kernel matrix from data matrices; that matrix
    should be an array of shape ``(n_samples, n_samples)``.
degree : int, default=3
    Degree of the polynomial kernel function ('poly').
    Ignored by all other kernels.
gamma : {'scale', 'auto'} or float, default='scale'
    Kernel coefficient for 'rbf', 'poly' and 'sigmoid'.
    - if ``gamma='scale'`` (default) is passed then it uses
      1 / (n_features * X.var()) as value of gamma,
    - if 'auto', uses 1 / n_features.
coef0 : float, default=0.0
    Independent term in kernel function.
    It is only significant in 'poly' and 'sigmoid'.
shrinking : bool, default=True
    Whether to use the shrinking heuristic.
probability : bool, default=False
    Whether to enable probability estimates. This must be enabled prior
    to calling `fit`, will slow down that method as it internally uses
    5-fold cross-validation, and `predict_proba` may be inconsistent with
    `predict`.
tol : float, default=1e-3
    Tolerance for stopping criterion.
cache_size : float, default=200
    Specify the size of the kernel cache (in MB).
class_weight : dict or 'balanced', default=None
    Set the parameter C of class i to class_weight[i]*C for
    SVC. If not given, all classes are supposed to have
    weight one.
    The "balanced" mode uses the values of y to automatically adjust
    weights inversely proportional to class frequencies in the input data
    as ``n_samples / (n_classes * np.bincount(y))``.
verbose : bool, default=False
    Enable verbose output. Note that this setting takes advantage of a
    per-process runtime setting in libsvm that, if enabled, may not work
    properly in a multithreaded context.
max_iter : int, default=-1
    Hard limit on iterations within solver, or -1 for no limit.
decision_function_shape : {'ovo', 'ovr'}, default='ovr'
    Whether to return a one-vs-rest ('ovr') decision function of shape
    (n_samples, n_classes) as all other classifiers, or the original
    one-vs-one ('ovo') decision function of libsvm which has shape
    (n_samples, n_classes * (n_classes - 1) / 2). However, note that
    internally, one-vs-one ('ovo') is always used as a multi-class strategy
    to train models; an ovr matrix is only constructed from the ovo matrix.
    The parameter is ignored for binary classification.
break_ties : bool, default=False
    If true, ``decision_function_shape='ovr'``, and number of classes > 2,
    :term:`predict` will break ties according to the confidence values of
    :term:`decision_function`; otherwise the first class among the tied
    classes is returned. Please note that breaking ties comes at a
    relatively high computational cost compared to a simple predict.
random_state : int, RandomState instance or None, default=None
    Controls the pseudo random number generation for shuffling the data for
    probability estimates. Ignored when `probability` is False.
    Pass an int for reproducible output across multiple function calls.
"""

Kernel = Literal["linear", "poly", "rbf", "sigmoid", "precomputed"]
Gamma = Union[Literal["scale", "auto"], float]
ClassWeight = Optional[dict | Literal["balanced"]]
DecisionFunctionShape = Literal["ovo", "ovr"]


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class SupportVectorClassifierConfig(ModelConfig):
    C: float = 1
    kernel: Kernel = "rbf"
    degree: int = 3
    gamma: Gamma = "scale"
    coef0: float = 0
    shrinking: bool = True
    probability: bool = False
    tol: float = 1e-3
    cache_size: float = 200
    class_weight: ClassWeight = "balanced"
    verbose: bool = False
    max_iter: int = -1
    decision_function_shape: DecisionFunctionShape = "ovr"
    break_ties: bool = False
    random_state: int | RandomState | None = None


class SupportVectorClassifier(
    SKLearnClassificationMixin,
    SKLearnModelBase[SupportVectorClassifierConfig],
):
    @property
    def _config_builder(self) -> type[SupportVectorClassifierConfig]:
        return SupportVectorClassifierConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> SVC:
        config = self.model_config
        return SVC(
            C=config.C,
            kernel=config.kernel,
            degree=config.degree,
            gamma=config.gamma,
            coef0=config.coef0,
            shrinking=config.shrinking,
            probability=config.probability,
            tol=config.tol,
            cache_size=config.cache_size,
            class_weight=config.class_weight,
            verbose=config.verbose,
            max_iter=config.max_iter,
            decision_function_shape=config.decision_function_shape,
            break_ties=config.break_ties,
            random_state=config.random_state,
        )

    def _predict_proba(self, data: datasets.Dataset, **kwargs: Any) -> PredictionResult:
        if not self.model_config.probability:
            raise RuntimeError(
                "Cannot use the predict_proba for this Support Vector Classifier, as the model config has"
                "probability=False. If you want to use predict_proba function, make sure to set probability=True in the model config before training the model.",
            )
        else:
            return super()._predict_proba(data, **kwargs)



import fsspec

from molflux.datasets.filesystems import S3FileSystem


def test_s3_filesystem_is_fsspec_compliant(
    fixture_mock_s3_client,
    fixture_mock_s3_filesystem,
):
    """That our custom S3FileSystem implements the fsspec protocol."""
    filesystem = S3FileSystem()
    assert isinstance(filesystem, fsspec.AbstractFileSystem)


def test_s3_filesystem_is_registered_with_fsspec(
    fixture_mock_s3_client,
    fixture_mock_s3_filesystem,
):
    """That our custom S3FileSystem is registered as an official fsspec entrypoint.

    References:
        https://filesystem-spec.readthedocs.io/en/latest/developer.html#implementing-a-backend
    """
    filesystem = fsspec.filesystem("s3")
    assert isinstance(filesystem, S3FileSystem)


def test_s3_path_gets_mapped_to_our_s3_filesystem(
    fixture_mock_s3_client,
    fixture_mock_s3_filesystem,
):
    """That fsspec dispatches s3 paths to our custom S3FileSystem."""
    path = "s3://my-bucket/test"
    fs_token_paths = fsspec.get_fs_token_paths(path)
    assert isinstance(fs_token_paths[0], S3FileSystem)



from collections.abc import Iterable
from os import PathLike as OSPathLike
from typing import Any, Union

RepresentationResult = dict[str, Any]
ArrayLike = Iterable[Any]
SmilesArray = Iterable[str]
OEMolArray = Iterable[Any]
MolArray = Union[SmilesArray, OEMolArray, Iterable[bytes]]

PathLike = Union[str, OSPathLike]
Fingerprint = list[int]



"""Uncertainty measure: Computes a given metric over a range of thresholded datasets based on uncertainty."""

import logging
from typing import Any

import evaluate
import numpy as np
from scipy.interpolate import interp1d
from tqdm.auto import tqdm

import datasets
from molflux.metrics import load_metric
from molflux.metrics.bases import UncertaintyMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_CITATION = """\
@article{leibig2017leveraging,
  title={Leveraging uncertainty information from deep neural networks for disease detection},
  author={Leibig, Christian and Allken, Vaneeda and Ayhan, Murat Seckin and Berens, Philipp and Wahl, Siegfried},
  journal={Scientific reports},
  volume={7},
  number={1},
  pages={1--14},
  year={2017},
  publisher={Nature Publishing Group}
}
"""

_DESCRIPTION = """\
Computes regression metrics over a range of uncertainty thresholds.
Measures the uncertainty quality, eg. we expect performance to get worse when including predictions with higher
uncertainty.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Estimated targets as returned by a classifier/regressor.
    references: Ground truth (correct) target values.
    standard_deviations: Estimated standard deviations for the predictions. The lower the value the more confident
        that the prediction is correct.
    metric_name (optional): The name of the metric, used to compute performance over the alternating thresholded
        dataset. All metrics from the catalogue are possible. Defaults to mean_squared_error.
    num_of_threshold_steps: The number of thresholds (unique values in `uncertainties`) to use for computing curve
        (evenly sampled from the threshold distribution). Used to speed up computation. (default = -1, use all thresholds)

Returns:
    uncertainty_based_rejection: Nested dictionary. For each threshold (unique values in uncertainties), there is one
        dictionary with an integer as key and a result dictionary as value:
                metric_name --> computed metric as requested via metric_name
                uncertainty_thr --> All samples with <= uncertainty value were included in the analysis.
                size_of_thresholded_ds --> Number of samples in the thresholded dataset.
                frac_of_retained_data --> Fraction of the thresholded dataset in comparison to the original dataset.

        If num_of_theshold_steps != -1, the number of elements of the dictionary will be less.

Examples:
    >>> from molflux.metrics import load_metric
    >>> m = load_metric("uncertainty_based_rejection")
    >>> ref = [0, 1, 2, 3, 4]
    >>> pred = [0, 2, 2, 3, 5]
    >>> standard_deviations = [0, 1, 0.2, 0.2, 1.5]
    >>> result = m.compute(references=ref, predictions=pred, standard_deviations=standard_deviations)
    >>> result["uncertainty_based_rejection"][1]
    {'mean_squared_error': 0.0, 'uncertainty_thr': 0.2..., 'size_of_thresholded_ds': 3, 'frac_of_retained_data': 0.6}
    >>> result = m.compute(references=ref, predictions=pred, standard_deviations=standard_deviations, metric_name='max_error')
    >>> result["uncertainty_based_rejection"][2]
    {'max_error': 1.0, 'uncertainty_thr': 1.0, 'size_of_thresholded_ds': 4, 'frac_of_retained_data': 0.8}

"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class UncertaintyBasedRejection(UncertaintyMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Value("float"),
                    "references": datasets.Value("float"),
                },
            ),
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        standard_deviations: ArrayLike | None = None,
        prediction_intervals: ArrayLike | None = None,
        metric_name: str = "mean_squared_error",
        progress_bar: bool = True,
        num_of_threshold_steps: int = -1,
        override_huggingface: bool = False,
        **kwargs: Any,
    ) -> MetricResult:
        """

        Args:
            predictions:
            references:
            uncertainties:
            metric_name:
            progress_bar:
            num_of_threshold_steps: number of thresholds to use for computing curve (evenly sampled from the threshold
            distribution) (default = -1, use all thresholds)
            override_huggingface: override the hugging face compute method by using the _compute of the metrics.
            makes computation faster, but might result in unstable behaviour!!!
            **kwargs:

        Returns:

        """

        if standard_deviations is not None:
            uncertainties = np.array(standard_deviations)
        elif prediction_intervals is not None:
            lower_bound, upper_bound = zip(*prediction_intervals, strict=False)
            uncertainties = np.array(upper_bound) - np.array(lower_bound)
        else:
            raise ValueError(
                "Please provide either uncertainties or prediction intervals.",
            )
        predictions = np.array(predictions)
        references = np.array(references)
        thresholds = list(np.unique(uncertainties))

        # if num_of_threshold_steps < len(thresholds), subsample thresholds by linear interpolation
        if (num_of_threshold_steps != -1) and (
            num_of_threshold_steps < len(thresholds)
        ):
            # interpolator
            threshold_interpolator = interp1d(
                np.linspace(0, 1, len(thresholds)),
                thresholds,
            )

            # subsample
            thresholds = threshold_interpolator(
                np.linspace(0, 1, num_of_threshold_steps),
            )

        molflux_metric = load_metric(metric_name)

        results_dic = {}
        for ii, thr in enumerate(
            tqdm(
                thresholds,
                "Computing uncertainty based rejection",
                disable=(not progress_bar),
            ),
        ):
            mask = uncertainties <= thr
            pred_filtered = predictions[mask]
            ref_filtered = references[mask]
            if len(ref_filtered) > 1:
                if override_huggingface:
                    results = molflux_metric._compute(  # type: ignore[attr-defined]
                        references=ref_filtered,
                        predictions=pred_filtered,
                    )
                else:
                    results = molflux_metric.compute(
                        references=ref_filtered,
                        predictions=pred_filtered,
                    )

                results["uncertainty_thr"] = thr
                results["size_of_thresholded_ds"] = len(pred_filtered)
                results["frac_of_retained_data"] = len(pred_filtered) / float(
                    len(predictions),
                )
                results_dic[ii] = results

        return {self.tag: results_dic}



from collections.abc import Iterable
from typing import Any

from molflux.modelzoo.catalogue import get_model_cls
from molflux.modelzoo.parsers import Spec, dict_parser, yaml_parser
from molflux.modelzoo.protocols import Model, Models
from molflux.modelzoo.typing import PathLike


def load_model(name: str, **init_kwargs: Any) -> Model:
    """Loads a `Model` from the catalogue."""

    # Fetch relevant model class from the catalogue
    model_cls = get_model_cls(model_name=name)

    # Instantiate model object
    model = model_cls(**init_kwargs)

    return model


def _load_from_spec(spec: Spec) -> Model:
    """Loads a model from a validated Spec."""

    # Build strategy
    model = load_model(name=spec.name, **spec.config)

    return model


def load_from_dict(dictionary: dict[str, Any]) -> Model:
    """Loads a model from a config dict."""

    # Validate dictionary
    spec = dict_parser(dictionary=dictionary)

    return _load_from_spec(spec=spec)


def load_from_dicts(dictionaries: Iterable[dict[str, Any]]) -> Models:
    """Loads models from an iterable of dicts."""

    models = (load_from_dict(dictionary) for dictionary in dictionaries)
    return {model.tag: model for model in models}


def load_from_yaml(path: PathLike) -> Models:
    """Loads models from a yaml config file."""

    specs = yaml_parser(path=path)

    models = (_load_from_spec(spec=spec) for spec in specs)
    return {model.tag: model for model in models}



from typing import Literal

from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.sklearn import SKLearnModelBase

try:
    from sklearn.ensemble import ExtraTreesRegressor as SKExtraTreesRegressor
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from None


_DESCRIPTION = """
This is an sklearn extra trees regressor model.

The extra trees regressor is a meta estimator that fits a number of randomized decision
trees (a.k.a. extra-trees) on various sub-samples of the dataset and uses averaging to
improve the predictive accuracy and control over-fitting.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
n_estimators : int, default=100
    The number of trees in the forest.
criterion : {"squared_error", "absolute_error", "friedman_mse", "poisson"}, \
        default="squared_error"
    The function to measure the quality of a split. Supported criteria
    are "squared_error" for the mean squared error, which is equal to
    variance reduction as feature selection criterion and minimizes the L2
    loss using the mean of each terminal node, "friedman_mse", which uses
    mean squared error with Friedman's improvement score for potential
    splits, "absolute_error" for the mean absolute error, which minimizes
    the L1 loss using the median of each terminal node, and "poisson" which
    uses reduction in Poisson deviance to find splits.
    Training using "absolute_error" is significantly slower
    than when using "squared_error".
max_depth : int, default=None
    The maximum depth of the tree. If None, then nodes are expanded until
    all leaves are pure or until all leaves contain less than
    min_samples_split samples.
min_samples_split : int or float, default=2
    The minimum number of samples required to split an internal node:
    - If int, then consider `min_samples_split` as the minimum number.
    - If float, then `min_samples_split` is a fraction and
      `ceil(min_samples_split * n_samples)` are the minimum
      number of samples for each split.
min_samples_leaf : int or float, default=1
    The minimum number of samples required to be at a leaf node.
    A split point at any depth will only be considered if it leaves at
    least ``min_samples_leaf`` training samples in each of the left and
    right branches.  This may have the effect of smoothing the model,
    especially in regression.
    - If int, then consider `min_samples_leaf` as the minimum number.
    - If float, then `min_samples_leaf` is a fraction and
      `ceil(min_samples_leaf * n_samples)` are the minimum
      number of samples for each node.
min_weight_fraction_leaf : float, default=0.0
    The minimum weighted fraction of the sum total of weights (of all
    the input samples) required to be at a leaf node. Samples have
    equal weight when sample_weight is not provided.
max_features : {"sqrt", "log2", None}, int or float, default=1.0
    The number of features to consider when looking for the best split:
    - If int, then consider `max_features` features at each split.
    - If float, then `max_features` is a fraction and
      `max(1, int(max_features * n_features_in_))` features are considered at each
      split.
    - If "sqrt", then `max_features=sqrt(n_features)`.
    - If "log2", then `max_features=log2(n_features)`.
    - If None or 1.0, then `max_features=n_features`.
    .. note::
        The default of 1.0 is equivalent to bagged trees and more
        randomness can be achieved by setting smaller values, e.g. 0.3.
    Note: the search for a split does not stop until at least one
    valid partition of the node samples is found, even if it requires to
    effectively inspect more than ``max_features`` features.
max_leaf_nodes : int, default=None
    Grow trees with ``max_leaf_nodes`` in best-first fashion.
    Best nodes are defined as relative reduction in impurity.
    If None then unlimited number of leaf nodes.
min_impurity_decrease : float, default=0.0
    A node will be split if this split induces a decrease of the impurity
    greater than or equal to this value.
    The weighted impurity decrease equation is the following::
        N_t / N * (impurity - N_t_R / N_t * right_impurity
                            - N_t_L / N_t * left_impurity)
    where ``N`` is the total number of samples, ``N_t`` is the number of
    samples at the current node, ``N_t_L`` is the number of samples in the
    left child, and ``N_t_R`` is the number of samples in the right child.
    ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
    if ``sample_weight`` is passed.
bootstrap : bool, default=False
    Whether bootstrap samples are used when building trees. If False, the
    whole dataset is used to build each tree.
oob_score : bool or callable, default=False
    Whether to use out-of-bag samples to estimate the generalization score.
    By default, :func:`~sklearn.metrics.r2_score` is used.
    Provide a callable with signature `metric(y_true, y_pred)` to use a
    custom metric. Only available if `bootstrap=True`.
n_jobs : int, default=None
    The number of jobs to run in parallel. :meth:`fit`, :meth:`predict`,
    :meth:`decision_path` and :meth:`apply` are all parallelized over the
    trees. ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
    context. ``-1`` means using all processors.
random_state : int or None, default=None
    Controls 3 sources of randomness:
    - the bootstrapping of the samples used when building trees
      (if ``bootstrap=True``)
    - the sampling of the features to consider when looking for the best
      split at each node (if ``max_features < n_features``)
    - the draw of the splits for each of the `max_features`
verbose : int, default=0
    Controls the verbosity when fitting and predicting.
warm_start : bool, default=False
    When set to ``True``, reuse the solution of the previous call to fit
    and add more estimators to the ensemble, otherwise, just fit a whole
    new forest.
ccp_alpha : non-negative float, default=0.0
    Complexity parameter used for Minimal Cost-Complexity Pruning. The
    subtree with the largest cost complexity that is smaller than
    ``ccp_alpha`` will be chosen. By default, no pruning is performed.
max_samples : int or float, default=None
    If bootstrap is True, the number of samples to draw from X
    to train each base estimator.
    - If None (default), then draw `X.shape[0]` samples.
    - If int, then draw `max_samples` samples.
    - If float, then draw `max_samples * X.shape[0]` samples. Thus,
      `max_samples` should be in the interval `(0.0, 1.0]`.
"""


Criterion = Literal["squared_error", "absolute_error", "friedman_mse", "poisson"]
MaxFeaturesCallable = Literal["sqrt", "log2"]


class Config:
    extra = "forbid"
    arbitrary_types_allowed = True
    smart_union = True


@dataclasses.dataclass(config=Config)
class ExtraTreesRegressorConfig(ModelConfig):
    n_estimators: int = 100
    criterion: Criterion = "squared_error"
    max_depth: int | None = None
    min_samples_split: int | float = 2
    min_samples_leaf: int | float = 1
    min_weight_fraction_leaf: float = 0.0
    max_features: float | MaxFeaturesCallable | None = "sqrt"
    max_leaf_nodes: int | None = None
    min_impurity_decrease: float = 0.0
    bootstrap: bool = False
    oob_score: bool = False
    n_jobs: int | None = None
    random_state: int | None = None
    verbose: int = 0
    warm_start: bool = False
    ccp_alpha: float = 0.0
    max_samples: int | float | None = None

    def __post_init_post_parse__(self) -> None:
        if self.ccp_alpha < 0:
            raise ValueError("Complexity parameter `ccp_alpha` cannot be negative.")


class ExtraTreesRegressor(SKLearnModelBase[ExtraTreesRegressorConfig]):
    @property
    def _config_builder(self) -> type[ExtraTreesRegressorConfig]:
        return ExtraTreesRegressorConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> SKExtraTreesRegressor:
        config = self.model_config
        return SKExtraTreesRegressor(
            n_estimators=config.n_estimators,
            criterion=config.criterion,
            max_depth=config.max_depth,
            min_samples_split=config.min_samples_split,
            min_samples_leaf=config.min_samples_leaf,
            min_weight_fraction_leaf=config.min_weight_fraction_leaf,
            max_features=config.max_features,
            max_leaf_nodes=config.max_leaf_nodes,
            min_impurity_decrease=config.min_impurity_decrease,
            bootstrap=config.bootstrap,
            oob_score=config.oob_score,
            n_jobs=config.n_jobs,
            random_state=config.random_state,
            verbose=config.verbose,
            warm_start=config.warm_start,
            ccp_alpha=config.ccp_alpha,
            max_samples=config.max_samples,
        )



"""
Tests for API features advertised as part of the Estimator protocol.
"""

from typing import Protocol

from molflux.modelzoo.protocols import Estimator


def test_class_is_protocol():
    """That is a protocol."""
    assert issubclass(type(Estimator), type(Protocol))


def test_has_init():
    """That can be initialised."""
    assert hasattr(Estimator, "__init__")
    assert callable(Estimator.__init__)


def test_has_metadata():
    """That has a metadata property."""
    assert hasattr(Estimator, "metadata")
    assert not callable(Estimator.metadata)


def test_has_name():
    """That has a name property."""
    assert hasattr(Estimator, "name")
    assert not callable(Estimator.name)


def test_has_tag():
    """That has a tag property."""
    assert hasattr(Estimator, "tag")
    assert not callable(Estimator.tag)


def test_has_x_features():
    """That has a x_features property."""
    assert hasattr(Estimator, "x_features")
    assert not callable(Estimator.x_features)


def test_has_y_features():
    """That has a y_features property."""
    assert hasattr(Estimator, "y_features")
    assert not callable(Estimator.y_features)


def test_has_predict():
    """That has a predict method."""
    assert hasattr(Estimator, "predict")
    assert callable(Estimator.predict)



from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.sklearn import SKLearnModelBase

try:
    from sklearn.cross_decomposition import PLSRegression as SKPLSRegression
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from None


_DESCRIPTION = """
Partial Least Squares regressor.

PLSRegression is also known as PLS1 (single targets) and PLS2 (multiple targets).

Much like Lasso, PLSRegression is a form of regularized linear regression where
the number of components controls the strength of the regularization.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
n_components : int, default=2
    Number of components to keep. Should be in `[1, min(n_samples,
    n_features, n_targets)]`.
scale : bool, default=True
    Whether to scale the input data.
max_iter : int, default=500
    The maximum number of iterations of the power method.
tol : float, default=1e-06
    The tolerance used as convergence criteria in the power method: the
    algorithm stops whenever the squared norm of `u_i - u_{i-1}` is less
    than `tol`, where `u` corresponds to the left singular vector.
inplace : bool, default=False
    Whether to copy the training data before applying centering,
    and potentially scaling. If `True`, these operations will be done
    inplace, modifying both arrays.
"""


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class PLSRegressorConfig(ModelConfig):
    n_components: int = 2
    scale: bool = True
    max_iter: int = 500
    tol: float = 1e-06
    inplace: bool = False


class PLSRegressor(SKLearnModelBase[PLSRegressorConfig]):
    @property
    def _config_builder(self) -> type[PLSRegressorConfig]:
        return PLSRegressorConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> SKPLSRegression:
        config = self.model_config
        return SKPLSRegression(
            n_components=config.n_components,
            scale=config.scale,
            max_iter=config.max_iter,
            tol=config.tol,
            copy=not config.inplace,
        )



from typing import Any

try:
    from rdkit.Chem import rdFingerprintGenerator

    from molflux.features.representations.rdkit._utils import rdkit_mol_from_smiles
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.rdkit._utils import to_smiles
from molflux.features.typing import Fingerprint, MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
Topological-torsion fingerprints, as described in:

The topological torsion (and related atom-pair) fingerprints, are implemented
based on the original papers. Atoms are typed based on atomic number,
number of pi electrons, and the degree of the atom. Optionally information
about atomic chirality can also be integrated into the atom types. Both
fingerprint types can be generated in explicit or sparse form and as bit or
count vectors. These fingerprint types are different from the others in the
RDKit in that bits in the sparse form of the fingerprint can be directly
explained (i.e. the “hashing function” used is fully reversible).

These fingerprints were originally “intended” to be used in count-vectors and
they seem to work better that way. The default behavior of the explicit
bit-vector forms of both fingerprints is to use a “count simulation” procedure
where multiple bits are set for a given feature if it occurs more than once.
The default behavior is to use 4 fingerprint bits for each feature
(so a 2048 bit fingerprint actually stores information about the same number of
features as a 512 bit fingerprint that isn`t using count simulation). The bins
correspond to counts of 1, 2, 4, and 8. As an example of how this works: if a
feature occurs 5 times in a molecule, the bits corresponding to counts 1, 2,
and 4 will be set.

R. Nilakantan, N. Bauman, J. S. Dixon, R. Venkataraghavan;
“Topological Torsion: A New Molecular Descriptor for SAR Applications.
Comparison with Other Descriptors” JCICS 27, 82-85 (1987).
"""


class TopologicalTorsion(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        include_chirality: bool = False,
        torsion_atom_count: int = 4,
        count_simulation: bool = True,
        count_bounds: object | None = None,
        fp_size: int = 2048,
        atom_invariants_generator: object | None = None,
        **kwargs: Any,
    ) -> dict[str, list[Fingerprint]]:
        """
        Generates a folded topological torsion fingerprint for each input molecule from a fingerprint generator.

        Args:
            samples: The molecules to be fingerprinted.
            include_chirality: if set, chirality will be used in the atom
                invariants, this is ignored if atom_invariants_generator is provided.
                Defaults to `False`.
            torsion_atom_count: the number of atoms to include in the "torsions".
                Default is 4.
            count_simulation:  if set, use count simulation while  generating
                the fingerprint. Defaults to `True`.
            count_bounds: boundaries for count simulation, corresponding bit
                will be  set if the count is higher than the number provided
                for that spot. Defaults to `None`.
            fp_size: size of the generated fingerprint, does not affect the
                sparse versions. Defaults is 2048.
            atom_invariants_generator: atom invariants to be used during
                fingerprint generation. Defaults to `None`.

        Returns:
            Folded atom-pair fingerprints, as lists of bits.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('topological_torsion')
            >>> samples = ['CN1C=NC2=C1C(=O)N(C(=O)N2C)C']
            >>> representation.featurise(samples, fp_size=16)
            {'topological_torsion': [[1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0]]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        ttgen = rdFingerprintGenerator.GetTopologicalTorsionGenerator(
            includeChirality=include_chirality,
            torsionAtomCount=torsion_atom_count,
            countSimulation=count_simulation,
            countBounds=count_bounds,
            fpSize=fp_size,
            atomInvariantsGenerator=atom_invariants_generator,
        )

        topological_torsion_fp_list: list[Fingerprint] = []
        for sample in samples:
            with featurisation_error_harness(sample):
                smiles = to_smiles(sample)
                mol = rdkit_mol_from_smiles(smiles)
                fingerprint = list(ttgen.GetFingerprint(mol))
                topological_torsion_fp_list.append(fingerprint)

        return {self.tag: topological_torsion_fp_list}



import numpy as np
import pytest

from molflux.metrics import Metric, list_metrics, load_metric
from molflux.metrics.uncertainty.coefficient_of_variation import CoefficientOfVariation


@pytest.fixture()
def fixture_metric():
    return load_metric("coefficient_of_variation")


def test_metric_in_catalogue():
    """That the metric is registered in the catalogue."""
    catalogue = list_metrics()
    all_metric_names = [name for names in catalogue.values() for name in names]
    assert "coefficient_of_variation" in all_metric_names


def test_metric_is_mapped_to_correct_class(fixture_metric):
    """That the catalogue name is mapped to the appropriate class."""
    metric = fixture_metric
    assert isinstance(metric, CoefficientOfVariation)


def test_implements_protocol(fixture_metric):
    """That the metric implements the public Metric protocol."""
    metric = fixture_metric
    assert isinstance(metric, Metric)


def test_prediction_intervals(fixture_metric):
    """That the metric computes properly for prediction intervals."""
    metric = fixture_metric
    predictions = [0, 0, 0, 0]
    references = [0.9, 0.1, 0.2, 0.3]
    lower_bound = [0.5, 0.5, 0.0, 0.0]
    upper_bound = [1.0, 1.0, 0.5, 0.5]
    prediction_intervals = list(zip(lower_bound, upper_bound, strict=False))
    result = metric.compute(
        predictions=predictions,
        references=references,
        prediction_intervals=prediction_intervals,
    )

    assert result["coefficient_of_variation"] >= 0
    assert result["coefficient_of_variation"] <= 1


def test_standard_deviations(fixture_metric):
    """That the metric computes properly for standard deviations."""
    metric = fixture_metric
    predictions = [0, 0, 0, 0]
    references = [0.9, 0.1, 0.2, 0.3]
    standard_deviations = [0.5, 1, 0.5, 1]
    result = metric.compute(
        predictions=predictions,
        references=references,
        standard_deviations=standard_deviations,
    )

    expected = 1 / 3
    assert np.isclose(result["coefficient_of_variation"], expected, atol=0.01)


def test_raise_error_no_prediction_interval(fixture_metric):
    """That omitting prediction interval raises error."""
    metric = fixture_metric
    predictions = [3, 0.0, 2]
    references = [3, -0.0, 1]
    # prediction_intervals argument omitted
    with pytest.raises(RuntimeError):
        metric.compute(predictions=predictions, references=references)



import numpy as np
import pytest

from molflux.metrics import Metric, list_metrics, load_metric
from molflux.metrics.regression.out_of_sample_r2 import OutOfSampleR2


@pytest.fixture()
def fixture_metric():
    return load_metric("out_of_sample_r2")


def test_metric_in_catalogue():
    """That the metric is registered in the catalogue."""
    catalogue = list_metrics()
    all_metric_names = [name for names in catalogue.values() for name in names]
    assert "out_of_sample_r2" in all_metric_names


def test_metric_is_mapped_to_correct_class(fixture_metric):
    """That the catalogue name is mapped to the appropriate class."""
    metric = fixture_metric
    assert isinstance(metric, OutOfSampleR2)


def test_implements_protocol(fixture_metric):
    """That the metric implements the public Metric protocol."""
    metric = fixture_metric
    assert isinstance(metric, Metric)


def test_default_compute(fixture_metric):
    """That default scoring."""
    metric = fixture_metric
    predictions = [2.5, 0.0, 2, 8]
    references = [3, -0.5, 2, 7]
    dummy_prediction = np.mean(predictions)
    result = metric.compute(
        predictions=predictions,
        references=references,
        dummy_prediction=dummy_prediction,
    )
    assert pytest.approx(result["out_of_sample_r2"], 0.01) == 0.948


def test_perfect_score(fixture_metric):
    """Test score on perfect predictions."""
    metric = fixture_metric
    predictions = [3, -0.5, 2, 7]
    references = [3, -0.5, 2, 7]
    dummy_prediction = np.mean(predictions)
    result = metric.compute(
        predictions=predictions,
        references=references,
        dummy_prediction=dummy_prediction,
    )
    assert result["out_of_sample_r2"] == 1


def test_constant_model(fixture_metric):
    """Test null score on a constant model that always predicts the expected value disregarding input features."""
    metric = fixture_metric
    predictions = [2, 2, 2]
    references = [1, 2, 3]
    dummy_prediction = 2
    result = metric.compute(
        predictions=predictions,
        references=references,
        dummy_prediction=dummy_prediction,
    )
    assert result["out_of_sample_r2"] == 0
